#!/usr/bin/env python
# filename: cluster.py


#
# Copyright (c) 2024 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import subprocess as sp
from typing import Iterable, Optional, Union

from ..bin import get_path as get_binary_path
from ..core.sequence import Sequence
from ..io import make_dir, to_fasta

__all__ = ["mmseqs_search"]


def mmseqs_search(
    query: Union[str, Sequence, Iterable[Sequence]],
    target: Union[str, Sequence, Iterable[Sequence]],
    output_path: str,
    temp_directory: Optional[str] = None,
    search_type: Optional[int] = None,
    max_seqs: Optional[int] = None,
    max_evalue: Optional[float] = None,
    sensitivity: Optional[float] = None,
    format_mode: Optional[int] = None,
    format_output: Optional[str] = None,
    verbosity: Optional[int] = None,
    threads: Optional[int] = None,
    split_memory_limit: Optional[str] = None,
    additional_cli_args: Optional[str] = None,
    mmseqs_bin: Optional[str] = None,
    log_to: Optional[str] = None,
    debug: bool = False,
) -> str:
    """
    Run MMseqs2_ easy-search.

    Parameters
    ----------
    query : Union[str, Sequence, Iterable[Sequence]]
        Query sequences. Can be either:
            - A path to a FASTA file
            - A path to an MMseqs2 database
            - A Sequence object
            - An iterable of Sequence objects

    target : Union[str, Sequence, Iterable[Sequence]]
        Target sequences. Can be either:
            - A path to a FASTA file
            - A path to an MMseqs2 database
            - A Sequence object
            - An iterable of Sequence objects

    output_path : str
        Path to save the output to. If the parent directory does not exist, it will be created.

    temp_directory : Optional[str]
        Path to save temporary files to. If not provided, the system temp directory will be used.

    search_type : int
        Search type. Options are:
          - 0: auto
          - 1: amino acid
          - 2: translated
          - 3: nucleotide
          - 4: translated nucleotide alignment
        If not provided, the default search type is 0 (auto).

    max_seqs : int
        Maximum number of matches to return for each query sequence. If not provided, the default is 300.

    max_evalue : float
        Only return matches below this E-value. If not provided, the default is 1.0e-3.

    sensitivity : float
        Sensitivity. 1.0 is fastest but least sensitive, 7.5 is slowest but most sensitive.
        If not provided, the default sensitivity is 5.7.

    format_mode : int
        Format mode. Options are:
          - 0: BLAST-TAB
          - 1: SAM
          - 2: BLAST-TAB + query/db length
          - 3: pretty HTML
          - 4: BLAST-TAB + column headers
        BLAST-TAB (0) and BLAST-TAB + column headers (4) support custom output formats using ``format_output``.
        If not provided, the default format mode is 0 (BLAST-TAB).

    format_output : str
        Comma separated list of output columns. Options are:
          - query
          - target
          - evalue
          - gapopen
          - pident
          - fident
          - nident
          - qstart
          - qend
          - qlen
          - tstart
          - tend
          - tlen
          - alnlen
          - raw
          - bits
          - cigar
          - qseq
          - tseq
          - qheader
          - theader
          - qaln
          - taln
          - qframe
          - mismatch
          - qcov
          - tcov
        Default is ``"query,target,fident,alnlen,mismatch,gapopen,qstart,qend,tstart,tend,evalue,bits"``

    verbosity : int
        Verbosity level. 0 is quiet, 1 is +errors, 2 is +warnings, 3 is +info.
        If not provided, the default verbosity level is 3 (+info).

    threads : int
        Number of threads to use. If not provided, the default is to use all available threads.

    split_memory_limit : str
        Memory limit for split searches. Should be a string like ``"10G"`` or ``"100M"``.
        If not provided, the limit is the default set in ``mmseqs2`` (all system memory).

    additional_cli_args : str
        Additional command line arguments (as a string) which will be passed directly to ``mmseqs easy-search``.

        .. warning::
            Providing command-line options as both arguments to this function call and as `additional_cli_args`
            (for example, setting `max_seqs` to 100 and also providing ``"--max-seqs 100"`` to `additional_cli_args`)
            will cause an error. Don't do that.

    mmseqs_bin : str
        Path to the MMseqs binary. If not provided, the MMseqs binary packaged with ``abutils`` will be used.

    debug : bool
        If True, print the command and stdout/stderr to the console.

    Returns
    -------
    output_path : str
        Path to the output file.


    .. _MMseqs2:
        https://github.com/soedinglab/MMseqs2
    """
    # process input data
    if not isinstance(query, str):
        if isinstance(query, Sequence):
            query = [query]
        query = to_fasta(query)
    if not isinstance(target, str):
        if isinstance(target, Sequence):
            target = [target]
        target = to_fasta(target)

    # set up output and temp directories
    make_dir(os.path.dirname(output_path))
    if temp_directory is None:
        temp_directory = "/tmp"

    # build mmseqs command
    if mmseqs_bin is None:
        mmseqs_bin = get_binary_path("mmseqs")
    mmseqs_cmd = f"{mmseqs_bin} easy-search {query} {target} {output_path} {temp_directory} --search-type {search_type}"
    if sensitivity is not None:
        mmseqs_cmd += f" -s {sensitivity}"
    if max_seqs is not None:
        mmseqs_cmd += f" --max-seqs {max_seqs}"
    if max_evalue is not None:
        mmseqs_cmd += f" -e {max_evalue}"
    if format_mode is not None:
        mmseqs_cmd += f" --format-mode {format_mode}"
    if format_output is not None:
        mmseqs_cmd += f" --format-output {format_output}"
    if verbosity is not None:
        mmseqs_cmd += f" -v {verbosity}"
    if threads is not None:
        mmseqs_cmd += f" --threads {threads}"
    if split_memory_limit is not None:
        mmseqs_cmd += f" --split-memory-limit {split_memory_limit}"
    if additional_cli_args is not None:
        mmseqs_cmd += f" {additional_cli_args}"

    # run MMseqs
    p = sp.Popen(mmseqs_cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise RuntimeError(f"MMseqs command failed: {stderr.decode('utf-8')}")
    if debug:
        print("STDOUT:\n", stdout.decode("utf-8"))
        print("\nSTDERR:\n", stderr.decode("utf-8"))
    if log_to is not None:
        log_to = os.path.abspath(log_to)
        make_dir(os.path.dirname(log_to))
        with open(log_to, "w") as f:
            f.write("STDOUT:\n-----\n")
            f.write(stdout.decode("utf-8"))
            f.write("\n\nSTDERR:\n-----\n")
            f.write(stderr.decode("utf-8"))
    return output_path



#!/usr/bin/python
# filename: preprocessing.py

#
# Copyright (c) 2024 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import datetime
import os
import subprocess as sp
import tempfile
import time
from typing import Iterable, Optional, Union

import polars as pl
from natsort import natsorted
from tqdm.auto import tqdm

from .. import Sequence
from ..bin import get_path as get_binary_path
from ..io import (
    concatenate_files,
    delete_files,
    from_polars,
    list_files,
    make_dir,
    rename_file,
    to_fasta,
    to_polars,
)
from ..tools import alignment, cluster

__all__ = ["deduplicate", "reduction", "merge_fastqs", "group_fastq_pairs"]


class FASTQFile:
    """
    Base class for processing paired FASTQ filenames.

    Parameters
    ----------
    file : str
        Path to a FASTQ file.

    """

    def __init__(self, file: str):
        self.file = file
        self.path = os.path.abspath(file)
        self.basename = os.path.basename(self.path)
        self.dir = os.path.dirname(self.path)
        self.filename = self.basename.rstrip(".gz").rstrip(".fastq").rstrip(".fq")

    def __eq__(self, other):
        return self.name == other.name


class IlluminaFile(FASTQFile):
    """
    Class for processing Illumina-formatted FASTQ filenames.
    Also compatible with Element's "legacy" filename structure.

    Parameters
    ----------
    file : str
        Path to a FASTQ file.

    """

    def __init__(self, file: str):
        super().__init__(file)
        self.schema = "illumina"

    @property
    def name(self):
        return "_".join(self.filename.split("_")[:-4])

    @property
    def number(self):
        return self.filename.split(".")[0].split("_")[-1]

    @property
    def read(self):
        return self.filename.split("_")[-2]

    @property
    def lane(self):
        return self.filename.split("_")[-3]

    @property
    def sample(self):
        return self.filename.split("_")[-4]


class ElementFile(FASTQFile):
    """
    Class for processing Element-formatted FASTQ filenames.

    Parameters
    ----------
    file : str
        Path to a FASTQ file.

    """

    def __init__(self, file: str):
        super().__init__(file)
        self.schema = "element"

    @property
    def name(self):
        return "_".join(self.filename.split("_")[:-1])

    @property
    def number(self):
        return ""

    @property
    def read(self):
        return self.filename.split("_")[-1]

    @property
    def lane(self):
        return ""

    @property
    def sample(self):
        return ""


class MergeGroup:
    """
    Group of FASTQ files to be merged.

    Parameters
    ----------
    name : str
        Name of the sample.

    files : Iterable[FASTQFile]
        Iterable of FASTQFile objects.

    """

    def __init__(self, name: str, files: Iterable[FASTQFile]):
        self.name = name
        self.files = files
        self.merged_file = None  # assigned by merge function

    def merge(
        self,
        merged_directory: str,
        log_directory: Optional[str] = None,
        format: str = "fastq",
        algo: str = "fastp",
        binary_path: Optional[str] = None,
        minimum_overlap: int = 30,
        allowed_mismatches: int = 5,
        allowed_mismatch_percent: float = 20.0,
        trim_adapters: bool = True,
        adapter_file: Optional[str] = None,
        quality_trim: bool = True,
        window_size: int = 4,
        quality_cutoff: int = 20,
        merge_args: Optional[str] = None,
        compress_output: bool = False,
        verbose: bool = False,
        debug: bool = False,
    ) -> str:
        groups = self._group_by_lane()
        n_groups = len(groups)
        if verbose:
            print(f"\n{self.name}")
            # print("-" * len(self.name))
            if n_groups > 1:
                groups = tqdm(
                    groups,
                    desc="  - merging lanes",
                    total=n_groups,
                    unit="lane",
                    leave=False,
                )
            else:
                print("  - merging paired FASTQs")

        # merge function
        if algo.lower() == "fastp":
            merge_func = merge_fastqs_fastp
        else:
            raise ValueError(f"Invalid merge algorithm: {algo}. Must be 'fastp'.")

        # merge files
        merged_files = []
        compress_suffix = ".gz" if compress_output else ""
        self.merged_file = os.path.join(
            merged_directory, f"{self.name}.{format.lower()}{compress_suffix}"
        )
        for group in groups:
            r1, r2 = natsorted(group, key=lambda x: x.read)
            # add name and compression suffix, if necessary
            name = self.name if n_groups == 1 else f"{self.name}_{r1.lane}"
            merged_path = os.path.join(
                merged_directory, f"{name}.{format.lower()}{compress_suffix}"
            )
            # merge
            merge_func(
                r1.path,
                r2.path,
                merged_path,
                name=name,
                log_directory=log_directory,
                trim_adapters=trim_adapters,
                quality_trim=quality_trim,
                window_size=window_size,
                quality_cutoff=quality_cutoff,
                minimum_overlap=minimum_overlap,
                allowed_mismatches=allowed_mismatches,
                allowed_mismatch_percent=allowed_mismatch_percent,
                adapter_file=adapter_file,
                binary_path=binary_path,
                additional_args=merge_args,
                debug=debug,
            )
            merged_files.append(merged_path)

        # concatenate merged files if dataset has multiple lanes
        if len(merged_files) > 1:
            if verbose:
                print("  - concatenating merged files")
            concatenate_files(merged_files, self.merged_file)
            delete_files(merged_files)
        else:
            rename_file(merged_files[0], self.merged_file)
        return self.merged_file

    def _group_by_lane(self):
        lane_dict = {}
        for f in self.files:
            if f.lane not in lane_dict:
                lane_dict[f.lane] = []
            lane_dict[f.lane].append(f)
        return [
            group for lane, group in natsorted(lane_dict.items(), key=lambda x: x[0])
        ]


def merge_fastqs(
    files: Union[str, Iterable],
    output_directory: str,
    output_format: str = "fastq",
    log_directory: Optional[str] = None,
    schema: str = "illumina",
    algo: str = "fastp",
    binary_path: Optional[str] = None,
    merge_args: Optional[str] = None,
    minimum_overlap: int = 30,
    allowed_mismatches: int = 5,
    allowed_mismatch_percent: float = 20.0,
    trim_adapters: bool = True,
    adapter_file: Optional[str] = None,
    quality_trim: bool = True,
    window_size: int = 4,
    quality_cutoff: int = 20,
    compress_output: bool = False,
    debug: bool = False,
    verbose: bool = False,
) -> Iterable[str]:
    """
    Merge paired-end fastq files.

    Parameters
    ----------
    files : Union[str, Iterable]
        Path to a directory containing paired-end fastq files, or an iterable of paired-end fastq files.

    output_directory : str
        Path to the directory in which to save the merged files.

    output_format : str, optional
        Output format. Must be 'fastq' or 'fasta'. Default is 'fastq'.

    log_directory : str, optional
        Path to the directory in which to save the fastp reports.
        If not provided, fastp reports will not be saved.

    schema : str, optional
        Schema of the file names. Can be 'illumina' or 'element'. Default is 'illumina'.

    algo : str, optional
        Algorithm to use for merging. Must be 'fastp'. Default is 'fastp'.

    binary_path : str, optional
        Path to the merge algorithm binary. If not provided, the binary packaged with abutils will be used.

    merge_args : str, optional
        Additional arguments (as a string) to pass to the merge function.

    minimum_overlap : int, optional
        Minimum overlap between reads. Default is 30.

    allowed_mismatches : int, optional
        Allowed mismatches between reads. Default is 5.

    allowed_mismatch_percent : float, optional
        Allowed mismatch percentage between reads. Default is 20.0.

    trim_adapters : bool, optional
        If True, trim adapters. Default is True.

    adapter_file : str, optional
        Path to a FASTA file containing adapter sequences. If not provided, the default
        adapter sequences (Illumina TruSeq) will be used.

    quality_trim : bool, optional
        If True, trim low-quality bases. Default is True.

    window_size : int, optional
        Sliding window size for quality trimming. Default is 4.

    quality_cutoff : int, optional
        Mean quality cutoff for quality trimming. Default is 20.

    compress_output : bool, optional
        If True, compress the output file using gzip. Default is False.

    debug : bool, optional
        If True, print debug output. Default is False.

    verbose : bool, optional
        If True, print verbose output. Default is False.

    Returns
    -------
    list
        A list of paths to the merged FASTA/Q files.

    """
    # process input files
    if isinstance(files, str):
        if not os.path.isdir(files):
            err = f"The supplied file path ({files}) does not exist or is not a directory."
            raise ValueError(err)
        files = list_files(files, extension=[".fastq", ".fq", ".fastq.gz", ".fq.gz"])
    if schema.lower() == "illumina":
        files = [IlluminaFile(f) for f in files]
    elif schema.lower() == "element":
        files = [ElementFile(f) for f in files]
    else:
        raise ValueError(f"Invalid schema: {schema}. Must be 'illumina' or 'element'.")
    # group files by sample
    file_pairs = group_fastq_pairs(files, verbose=verbose)
    if verbose:
        print("--------------------")
        print("    MERGE FASTQs    ")
        print("--------------------")
        file_pairs = tqdm(file_pairs, leave=True)
    # merge files
    make_dir(output_directory)
    merged_files = []
    for fp in file_pairs:
        merged_file = fp.merge(
            merged_directory=output_directory,
            format=output_format,
            algo=algo,
            binary_path=binary_path,
            log_directory=log_directory,
            minimum_overlap=minimum_overlap,
            allowed_mismatches=allowed_mismatches,
            allowed_mismatch_percent=allowed_mismatch_percent,
            trim_adapters=trim_adapters,
            adapter_file=adapter_file,
            quality_trim=quality_trim,
            window_size=window_size,
            quality_cutoff=quality_cutoff,
            merge_args=merge_args,
            compress_output=compress_output,
            debug=debug,
            verbose=verbose,
        )
        merged_files.append(merged_file)
    return merged_files


def group_fastq_pairs(
    files: Iterable[FASTQFile], verbose: bool = False
) -> Iterable[MergeGroup]:
    """
    Group paired-end fastq files by sample name. If a sample has multiple lanes, the files will be combined.

    Parameters
    ----------
    files : Union[str, Iterable]
        Path to a directory containing paired-end fastq files, or an iterable of paired-end fastq files.

    verbose : bool, optional
        If True, print verbose output. Default is False.

    Returns
    -------
    list
        A list of MergeGroup objects.

    """
    group_dict = {}
    for f in files:
        if f.name not in group_dict:
            group_dict[f.name] = []
        group_dict[f.name].append(f)
    groups = [MergeGroup(name, group_files) for name, group_files in group_dict.items()]
    return natsorted(groups, key=lambda x: x.name)


def merge_fastqs_vsearch(
    forward: str,
    reverse: str,
    merged_file: str,
    output_format: str = "fasta",
    binary_path: Optional[str] = None,
    minimum_overlap: int = 30,
    allowed_mismatches: int = 5,
    allowed_mismatch_percent: float = 20.0,
    additional_args: Optional[str] = None,
    debug: bool = False,
) -> str:
    """
    Merge paired-end reads using vsearch.

    Parameters
    ----------
    forward : str
        Path to the forward read file.

    reverse : str
        Path to the reverse read file.

    merged_file : str
        Path to the merged read file.

    output_format : str, optional
        Output format. Must be 'fasta' or 'fastq'. Default is 'fasta'.

    binary_path : str, optional
        Path to a vsearch binary. If not provided, the vsearch binary packaged with abutils will be used.

    additional_args : str, optional
        Additional arguments (as a string) to pass to vsearch.

    debug : bool, optional
        If True, print vsearch stdout and stderr. Default is False.

    Returns
    -------
    str
        Path to the merged read file.

    """
    # validate input files
    if not os.path.isfile(forward):
        err = f"The supplied forward read file path ({forward}) does not exist or is not a file."
        raise ValueError(err)
    if not os.path.isfile(reverse):
        err = f"The supplied reverse read file path ({reverse}) does not exist or is not a file."
        raise ValueError(err)
    # make output directory
    out_dir = os.path.dirname(merged_file)
    make_dir(out_dir)
    # get the vsearch binary
    if binary_path is None:
        binary_path = get_binary_path("vsearch")
    # compile the vsearch command
    cmd = f"{binary_path} --fastq_mergepairs {forward} --reverse {reverse}"
    if output_format.lower() == "fasta":
        cmd += f" --fastaout {merged_file}"
    elif output_format.lower() == "fastq":
        cmd += f" --fastqout {merged_file}"
    else:
        err = f"Invalid output format: {output_format}. Must be 'fasta' or 'fastq'."
        raise ValueError(err)
    if additional_args is not None:
        cmd += f" {additional_args}"
    # merge reads
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if debug:
        print(stdout.decode())
        print(stderr.decode())
    if p.returncode != 0:
        err = f"Error merging reads with vsearch: {stderr.decode()}"
        raise ValueError(err)
    return merged_file


def merge_fastqs_fastp(
    forward: str,
    reverse: str,
    merged: str,
    binary_path: Optional[str] = None,
    minimum_overlap: int = 30,
    allowed_mismatches: int = 5,
    allowed_mismatch_percent: int = 20,
    correct_overlap_region: bool = False,
    trim_adapters: bool = True,
    adapter_file: Optional[str] = None,
    quality_trim: bool = True,
    window_size: int = 4,
    quality_cutoff: int = 20,
    name: Optional[str] = None,
    log_directory: Optional[str] = None,
    additional_args: Optional[str] = None,
    debug: bool = False,
) -> str:
    """
    Merge paired-end reads using vsearch.

    Parameters
    ----------
    forward : str
        Path to the forward read file.

    reverse : str
        Path to the reverse read file.

    merged : str
        Path to the merged read file. If the merged file ends in ".gz", the merged file
        will be gzip compressed.

    binary_path : str, optional
        Path to a fastp binary. If not provided, the fastp binary packaged with abutils will be used.

    minimum_overlap : int, optional
        Minimum overlap between reads. Default is 30.

    allowed_mismatches : int, optional
        Allowed mismatches between reads. Default is 5.

    allowed_mismatch_percent : float, optional
        Allowed mismatch percentage between reads. Default is 20.0.

    correct_overlap_region : bool, optional
        If True, correct the overlap region. Default is False.

    trim_adapters : bool, optional
        If True, trim adapters. Default is True.

    adapter_file : str, optional
        Path to a FASTA file containing adapter sequences. If not provided, the default
        adapter sequences (Illumina TruSeq) will be used.

    quality_trim : bool, optional
        If True, trim low-quality bases. Default is True.

    window_size : int, optional
        Sliding window size for quality trimming. Default is 4.

    quality_cutoff : int, optional
        Mean quality cutoff for quality trimming. Default is 20.

    name : str, optional
        Name of the sample. If not provided, the name of the merged file will be used.

    log_directory : str, optional
        Path to the directory in which to save the fastp report.
        If not provided, fastp reports will not be saved.

    additional_args : str, optional
        Additional arguments (as a string) to pass to fastp.

    debug : bool, optional
        If True, print vsearch stdout and stderr. Default is False.

    Returns
    -------
    str
        Path to the merged read file.

    """
    # validate input files
    if not os.path.isfile(forward):
        err = f"The supplied forward read file path ({forward}) does not exist or is not a file."
        raise ValueError(err)
    if not os.path.isfile(reverse):
        err = f"The supplied reverse read file path ({reverse}) does not exist or is not a file."
        raise ValueError(err)

    # make output directory
    merged = os.path.abspath(merged)
    out_dir = os.path.dirname(merged)
    make_dir(out_dir)

    # get the vsearch binary
    if binary_path is None:
        binary_path = get_binary_path("fastp")

    # compile the vsearch command
    cmd = f"{binary_path} -i '{forward}' -I '{reverse}' --merge --merged_out '{merged}'"
    cmd += f" --overlap_len_require {int(minimum_overlap)}"
    cmd += f" --overlap_diff_limit {int(allowed_mismatches)}"
    cmd += f" --overlap_diff_percent_limit {int(allowed_mismatch_percent)}"
    if correct_overlap_region:
        cmd += " --correction"

    # adapters
    if trim_adapters:
        if adapter_file is None:
            cmd += " --adapter_sequence=AGATCGGAAGAGCACACGTCTGAACTCCAGTCA"
            cmd += " --adapter_sequence_r2=AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGT"
        else:
            cmd += f" --adapter_fasta {adapter_file}"
    else:
        cmd += " --disable_adapter_trimming"

    # quality
    if quality_trim:
        cmd += " --cut_tail"
        cmd += f" --cut_tail_window_size {int(window_size)}"
        cmd += f" --cut_tail_mean_quality {int(quality_cutoff)}"
    else:
        cmd += " --disable_quality_filtering"

    # log
    if log_directory is None:
        log_directory = tempfile.mkdtemp()
    make_dir(log_directory)
    if name is None:
        if merged.endswith(".gz"):
            name = ".".join(os.path.basename(merged).split(".")[:-2])
        else:
            name = ".".join(os.path.basename(merged).split(".")[:-1])
    cmd += f" --html '{os.path.join(log_directory, name)}_fastp-report.html'"
    cmd += f" --json '{os.path.join(log_directory, name)}_fastp-report.json'"

    # additional CLI args
    if additional_args is not None:
        cmd += f" {additional_args}"

    # merge reads
    p = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if debug:
        print(stdout.decode())
        print(stderr.decode())
    if p.returncode != 0:
        err = f"Error merging reads with fastp: {stderr.decode()}"
        raise ValueError(err)
    return merged


#########################################
#                                       #
#         Helper functions              #
#                                       #
#########################################


def _deduplicate_sequences(
    df: pl.DataFrame, group_by_cols: list, keep_read_numbers: bool
):
    if not keep_read_numbers:
        return df.unique(subset=group_by_cols).sort("sequence")
    return (
        df.group_by(group_by_cols)
        .agg(
            pl.count().alias("count"),
            pl.col("sequence_id").first(),
        )
        .sort("sequence")
    )


def _write_fasta(
    df: pl.DataFrame,
    file_path: str,
    keep_read_numbers: bool,
    read_number_separator: str,
):
    if keep_read_numbers:
        df = df.with_columns(
            (
                pl.col("sequence_id")
                + read_number_separator
                + pl.col("count").cast(pl.Utf8)
            ).alias("sequence_id")
        )
    tuples_list = list(zip(df["sequence_id"].to_list(), df["sequence"].to_list()))
    to_fasta(tuples_list, file_path)
    print(f"Output written to FASTA file: {file_path}\n")


def _process_chains(
    step: str,
    chains: pl.DataFrame,
    min_cluster_size: int,
    clustering_threshold: float,
    consentroid: str,
    cluster_sizes_separator: str,
    keep_cluster_sizes: bool,
    output_format: str,
    debug: bool,
):
    """Internal function to generate a consentroid from a bin of sequences"""

    consentroids = []
    for vj in tqdm(
        chains["vj_bin"].unique().to_list(), desc=f"{step} chains processing"
    ):
        group = chains.filter(pl.col("vj_bin") == vj)
        if group.height < min_cluster_size:
            continue

        group_sequences = from_polars(group)
        clusters = cluster.cluster(group_sequences, threshold=clustering_threshold)
        for c in clusters:
            if c.size < min_cluster_size:
                continue
            if consentroid == "centroid":
                seq_id = (
                    f"{c.centroid.id}{cluster_sizes_separator}{c.size}"
                    if keep_cluster_sizes
                    else c.centroid.id
                )
                consentroids.append(Sequence(c.centroid.sequence, id=seq_id))
            elif consentroid == "consensus":
                _consensus = alignment.make_consensus(
                    [s for s in c.sequences],
                    algo="mafft",
                    alignment_kwargs={"mafft_bin": "mafft", "threads": 1},
                    debug=debug,
                )
                seq_id = (
                    f"{_consensus.id}{cluster_sizes_separator}{c.size}"
                    if keep_cluster_sizes
                    else _consensus.id
                )
                consentroids.append(Sequence(_consensus.sequence, id=seq_id))

        if debug:
            print(
                f"\tProcessed {len(clusters)} clusters for {vj}, corresponding to {len(group_sequences)} sequences"
            )

    if debug:
        print(f"\tTotal concentroids generated: {len(consentroids)}\n")

    if output_format == "fasta":
        return consentroids

    elif output_format == "airr":
        df = to_polars(consentroids)
        if keep_cluster_sizes and len(df) != 0:
            df = (
                df.with_columns(
                    pl.col("sequence_id")
                    .str.split_exact(
                        cluster_sizes_separator, 1
                    )  # Split into a struct with two fields
                    .alias("split_id")
                )
                .with_columns(
                    pl.col("split_id")
                    .struct.field("field_0")
                    .alias("sequence_id"),  # Extract the first field
                    pl.col("split_id")
                    .struct.field("field_1")
                    .cast(pl.Utf8)
                    .alias("count"),  # Extract the second field
                )
                .drop("split_id")
            )  # Drop the intermediate "split_id" column
        return df


def _process_chain_group(
    chains: pl.DataFrame,
    chain_type: str,
    umi: bool,
    min_cluster_size: int,
    clustering_threshold: float,
    consentroid: str,
    cluster_sizes_separator: str,
    keep_cluster_sizes: bool,
    output_format: str,
    debug: bool,
):
    # Generate vj_bin column
    if umi:
        chains = chains.with_columns(
            (pl.col("v_gene") + "_" + pl.col("j_gene") + "_" + pl.col("umi")).alias(
                "vj_bin"
            )
        )
    else:
        chains = chains.with_columns(
            (pl.col("v_gene") + "_" + pl.col("j_gene")).alias("vj_bin")
        )

    return _process_chains(
        chain_type,
        chains,
        min_cluster_size,
        clustering_threshold,
        consentroid,
        cluster_sizes_separator,
        keep_cluster_sizes,
        output_format,
        debug,
    )


#########################################
#                                       #
#           Main functions              #
#                                       #
#########################################


def deduplicate(
    project_folder: str,
    output: Optional[str] = None,
    output_format: str = "fasta",
    pool: bool = True,
    umi: bool = False,
    keep_read_numbers: bool = False,
    read_number_separator: str = "|",
    large_files: bool = False,
    debug: bool = False,
) -> None:
    """
    A polyvalent tool for deduplication of assigned reads.
    This function takes as input the AIRR-compliant
    tables and is specifically designed to handle extremely
    large files with a minimal footprint.

    Parameters
    ----------
    project_folder : str
        Path to the project folder containing AIRR-compliant tables.

    output : str, optional
        Subdirectory for output files. Created if non-existent. Defaults to None.

    output_format : str
        Either "fasta" or "airr". Default is "fasta".

    pool : bool, optional
        If True, pool all samples together. Defaults to True.

    umi : bool, optional
        If True, use UMI for deduplication. Defaults to False.

    keep_read_numbers : bool, optional
        If True, read numbers will be added to sequence names. Defaults to False.

    read_number_separator : str, optional
        Separator for read numbers in sequence names. Defaults to "|".

    large_files : bool, optional
        If True, optimize for large files (>100Go). Defaults to False.

    debug : bool, optional
        If True, print debug information. Defaults to False.

    Returns
    -------
    None. A FASTA- or TSV-formatted file is written to disk.
    """

    start = time.time()

    # Assert that output_format is valid
    if output_format not in ["fasta", "airr"]:
        raise ValueError("Invalid output_format. Must be 'fasta' or 'airr'.")

    # Preparing input and output file(s) / folder(s)
    if os.path.isfile(project_folder):
        project_folder = os.path.dirname(project_folder)
        if debug:
            print(f"Project folder is set to: {project_folder}")

    files = sorted(
        [
            f
            for f in list_files(
                project_folder,
                extension="tsv",
                recursive=True,
            )
            if "tmp" not in f
        ]
    )

    if len(files) == 0:
        print("No files found. Exiting.")
    elif len(files) == 1:
        pool = False

    if output:
        project_folder = os.path.join(project_folder, output)
        make_dir(project_folder)

    # Running main deduplication process
    total_sequences = 0
    pooled = []

    for file in files:
        sample = os.path.basename(file).split(".")[0]

        print(f"Processing {sample}")
        print("-" * (len(sample) + 11) + "\n")

        # Loading files
        if umi:
            try:
                df = pl.read_csv(
                    file,
                    columns=["sequence_id", "sequence", "umi"],
                    separator="\t",
                    null_values="None",
                    low_memory=True if large_files else False,
                )
            except Exception as e:
                print(f"Error reading file {file}: {e}")
                continue
        else:
            try:
                df = pl.read_csv(
                    file,
                    columns=["sequence_id", "sequence"],
                    separator="\t",
                    null_values="None",
                    low_memory=True if large_files else False,
                )
            except Exception as e:
                print(f"Error reading file {file}: {e}")
                continue

        print(f"Loaded {df.height:,} annotated sequences")
        total_sequences += df.shape[0]

        # Processing with deduplication in the presence of UMIs
        if umi:
            if debug:
                print(
                    f"Deduplicating with UMI ({df.unique(subset=['umi']).height} unique UMIs)"
                )

            df_unique = _deduplicate_sequences(
                df, ["sequence", "umi"], keep_read_numbers
            )

        # Processing with deduplication in the absence of UMIs
        else:
            df_unique = _deduplicate_sequences(df, ["sequence"], keep_read_numbers)

        msg = f"Found {df_unique.height:,} unique sequences"
        if pool:
            msg += " added to the pool\n"
        print(msg)

        if pool:
            pooled.append(df_unique)
        else:
            # Saving deduplicated single output to FASTA file
            if output_format == "fasta":
                fasta_file = os.path.join(project_folder, sample) + ".fasta"

                _write_fasta(
                    df_unique, fasta_file, keep_read_numbers, read_number_separator
                )

            # Saving deduplicated single output to TSV file
            elif output_format == "airr":
                tsv_file = os.path.join(project_folder, sample) + ".tsv"
                # Convert df_unique to LazyFrame for more efficient processing
                df_unique_lazy = pl.LazyFrame(df_unique)
                # Scan the complete original AIRR table
                new_df = pl.scan_csv(
                    file, separator="\t", low_memory=True if large_files else False
                )

                # Filter new_df to only include rows with matching sequence_ids
                filtered_df = new_df.join(
                    df_unique_lazy.select(["sequence_id"]), on="sequence_id", how="semi"
                ).select(new_df.columns)

                # Add the 'count' column to the filtered DataFrame (only for matching rows)
                if keep_read_numbers:
                    filtered_df = (
                        filtered_df.join(
                            df_unique_lazy.select(["sequence_id", "count"]),
                            on="sequence_id",
                            how="left",
                        )
                        .with_columns(
                            pl.col("count").fill_null(
                                0
                            )  # Replace null values with 0 if necessary
                        )
                        .rename(
                            {"count": "duplicates"}  # Rename the column
                        )
                    )

                # Sink the resulting DataFrame to a TSV file
                filtered_df.collect().write_csv(tsv_file, separator="\t")
                print(f"Output written to TSV file: {tsv_file}\n")

    if pool:
        pool_df = pl.concat(pooled)

        if not keep_read_numbers:
            pool_unique = pool_df.unique(subset=["sequence"]).sort("sequence")
        else:
            pool_unique = (
                pool_df.group_by("sequence")
                .agg(
                    pl.count().alias("count"),
                    pl.col("sequence_id").first(),
                )
                .sort("sequence")
            )

        print(f"\nFound {pool_unique.height:,} unique sequences in pooled data")

        # Saving deduplicated pooled output to FASTA file
        if output_format == "fasta":
            fasta_file = os.path.join(project_folder, "deduplicated_pool.fasta")
            _write_fasta(
                pool_unique, fasta_file, keep_read_numbers, read_number_separator
            )

        # Saving deduplicated pooled output to TSV file
        elif output_format == "airr":
            tsv_file = os.path.join(project_folder, "deduplicated_pool.tsv")

            # Convert df_unique to LazyFrame for more efficient processing
            df_unique_lazy = pl.LazyFrame(pool_unique)

            # Scan and concatenate multiple AIRR tables
            new_df = pl.concat(
                [
                    pl.scan_csv(
                        file, separator="\t", low_memory=True if large_files else False
                    )
                    for file in files
                ]
            )

            # Filter new_df to only include rows with matching sequence_ids
            filtered_df = new_df.join(
                df_unique_lazy.select(["sequence_id"]), on="sequence_id", how="semi"
            ).select(new_df.columns)

            # Add the 'count' column to the filtered DataFrame (only for matching rows)
            if keep_read_numbers:
                filtered_df = (
                    filtered_df.join(
                        df_unique_lazy.select(["sequence_id", "count"]),
                        on="sequence_id",
                        how="left",
                    )
                    .with_columns(
                        pl.col("count").fill_null(
                            0
                        )  # Replace null values with 0 if necessary
                    )
                    .rename(
                        {"count": "duplicates"}  # Rename the column
                    )
                )

            # Sink the resulting DataFrame to a TSV file
            filtered_df.collect().write_csv(tsv_file, separator="\t")
            print(f"Output written to TSV file: {tsv_file}\n")

    # Finalizing
    end = time.time()
    duration = end - start
    print(
        f"Deduplication complete. Total elapsed time: {datetime.timedelta(seconds=int(duration))}."
    )

    return


def reduction(
    project_folder: str,
    output: Optional[str] = None,
    output_format: str = "fasta",
    pool: bool = True,
    umi: bool = False,
    keep_cluster_sizes: bool = False,
    cluster_sizes_separator: str = "|",
    min_cluster_size: int = 3,
    clustering_threshold: float = 0.975,
    consentroid: str = "centroid",
    large_files: bool = False,
    debug: bool = False,
) -> None:
    """
    This function takes as an input AIRR-compliant tables (tsv) and proceeds to
    data reduction by clustering sequences to a high identity threshold.

    This is specifically designed to handle large files with minimal footprint.
    Preclustering can be applied to increase performance over large datasets.

    Parameters:
    project_folder (str): Path to the project folder containing AIRR-compliant tables.
    output (str, optional): Subdirectory for output files. Created if non-existent. Defaults to None.
    output_format (str): Either "fasta" or "airr". Default is "fasta".
    pool (bool, optional): If True, pool all samples together. Defaults to True.
    umi (bool, optional): If True, use UMI for clustering. Defaults to False.
    keep_cluster_sizes (bool, optional): If True, cluster sizes will be added to sequence names. Defaults to False.
    cluster_sizes_separator (str, optional): Separator for cluster sizes in sequence names. Defaults to '|'.
    min_cluster_size (int, optional): Minimum cluster size to consider. Defaults to 3.
    clustering_threshold (float, optional): Identity threshold for clustering. Defaults to 0.975.
    consentroid (str, optional): Method to determine cluster representative ('centroid' or 'consensus'). Defaults to 'centroid'.
    large_files (bool, optional): If True, optimize for large files (>100Go). Defaults to False.
    debug (bool, optional): If True, print debug information. Defaults to False.

    Returns: None. A fasta file is written to disk.
    """

    start = time.time()

    if umi:
        cluster_sizes_separator += "umi_count="
    else:
        cluster_sizes_separator += "cluster_size="

    # Assert that output_format is valid
    if output_format not in ["fasta", "airr"]:
        raise ValueError("Invalid output_format. Must be 'fasta' or 'airr'.")

    # Assert that clustering_threshold is valid
    if not (0 < clustering_threshold <= 1):
        raise ValueError("Clustering_threshold must be between 0 and 1.")

    # Assert that consentroid is valid
    if consentroid not in ["centroid", "consensus"]:
        raise ValueError("Consentroid must be either 'centroid' or 'consensus'.")

    # Check for incompatible consentroid and output_format arguments
    if consentroid == "consensus" and output_format == "airr":
        raise ValueError(
            "The 'consensus' consentroid method is incompatible with the 'airr' output format. "
            "Please use 'centroid' or choose 'fasta' as the output format."
        )

    # Preparing input and output file(s) / folder(s)
    if os.path.isfile(project_folder):
        project_folder = os.path.dirname(project_folder)
        if debug:
            print(f"Project folder is set to: {project_folder}")

    files = sorted(
        [
            f
            for f in list_files(
                project_folder,
                extension="tsv",
                recursive=True,
            )
            if not "tmp" in f
        ]
    )

    if len(files) == 0:
        print("No files found. Exiting.")
    elif len(files) == 1:
        pool = False

    if output:
        project_folder = os.path.join(project_folder, output)
        make_dir(project_folder)

    total_sequences = 0

    if pool:
        pooled_heavies = []
        pooled_lights = []

    # Processing files individually...
    for file in files:
        sample = os.path.basename(file).split(".")[0]

        print(f"Processing {sample}")
        print("-" * (len(sample) + 11) + "\n")

        keys = ["sequence_id", "v_gene", "j_gene", "locus", "sequence"]
        if umi:
            keys.append("umi")

        df = pl.read_csv(
            file,
            columns=keys,
            separator="\t",
            null_values="None",
            low_memory=True if large_files else False,
        )

        if not all(col in df.columns for col in keys):
            print(f"File {file} is missing required columns. Skipping...")
            continue

        print(f"Loaded {df.height:,} annotated sequences")
        total_sequences += df.height

        heavies = df.filter(pl.col("locus") == "IGH")
        lights = df.filter(pl.col("locus") != "IGH")

        if pool:
            pooled_heavies.append(heavies)
            pooled_lights.append(lights)

        else:
            sample_consentroids = []

            # Process individual heavy chains
            heavy_consentroids = _process_chain_group(
                heavies,
                "Heavy",
                umi,
                min_cluster_size,
                clustering_threshold,
                consentroid,
                cluster_sizes_separator,
                keep_cluster_sizes,
                output_format,
                debug,
            )
            sample_consentroids.extend(heavy_consentroids)

            # Process individual light chains
            light_consentroids = _process_chain_group(
                lights,
                "Lights",
                umi,
                min_cluster_size,
                clustering_threshold,
                consentroid,
                cluster_sizes_separator,
                keep_cluster_sizes,
                output_format,
                debug,
            )
            sample_consentroids.extend(light_consentroids)

            # Skipping file creation if no concentroids have been generated
            if not sample_consentroids:
                print(f"No clusters found for sample {sample}. Skipping output.\n")
                continue

            if output_format == "fasta":
                # Save consentroids to FASTA file
                fasta_file = os.path.join(project_folder, sample) + "_reduced.fasta"
                to_fasta(sample_consentroids, fasta_file)

            elif output_format == "airr":
                # Saving centroids to TSV file (consensus can't be exported as AIRR-annotated TSV table as they haven't been annotated yet)
                tsv_file = os.path.join(project_folder, sample + "_reduced.tsv")

                # Convert centroids to LazyFrame for more efficient processing
                centroids_lazy = pl.LazyFrame(
                    sample_consentroids,
                )

                # Scan the complete original AIRR table
                new_df = pl.scan_csv(
                    file, separator="\t", low_memory=True if large_files else False
                )

                # Filter new_df to only include rows with matching sequence_ids
                filtered_df = new_df.join(
                    centroids_lazy.select(["sequence_id"]), on="sequence_id", how="semi"
                ).select(new_df.columns)

                # Add the 'count' column to the filtered DataFrame (only for matching rows)
                if keep_cluster_sizes:
                    filtered_df = (
                        filtered_df.join(
                            centroids_lazy.select(["sequence_id", "count"]),
                            on="sequence_id",
                            how="left",
                        )
                        .with_columns(
                            pl.col("count").fill_null(
                                0
                            )  # Replace null values with 0 if necessary
                        )
                        .rename(
                            {
                                "count": "umi_count" if umi else "cluster_size"
                            }  # Rename the column
                        )
                    )

                # Sink the resulting DataFrame to a TSV file
                filtered_df.collect().write_csv(tsv_file, separator="\t")
                print(f"Output written to TSV file: {tsv_file}\n")

    if pool:
        heavies = pl.concat(pooled_heavies)
        lights = pl.concat(pooled_lights)

        all_consentroids = []

        # Processing pooled heavy chains
        heavy_consentroids = _process_chain_group(
            heavies,
            "Heavy",
            umi,
            min_cluster_size,
            clustering_threshold,
            consentroid,
            cluster_sizes_separator,
            keep_cluster_sizes,
            output_format,
            debug,
        )
        all_consentroids.extend(heavy_consentroids)

        # Processing pooled light chains
        light_consentroids = _process_chain_group(
            lights,
            "Lights",
            umi,
            min_cluster_size,
            clustering_threshold,
            consentroid,
            cluster_sizes_separator,
            keep_cluster_sizes,
            output_format,
            debug,
        )
        all_consentroids.extend(light_consentroids)

        if output_format == "fasta":
            # Saving pooled consentroids to FASTA file
            fasta_file = os.path.join(project_folder, "reduced_pool.fasta")
            to_fasta(all_consentroids, fasta_file)

        elif output_format == "airr":
            # Saving consentroids to TSV file
            tsv_file = os.path.join(project_folder, "reduced_pool.tsv")

            # Convert centroids to LazyFrame for more efficient processing
            centroids_lazy = pl.LazyFrame(all_consentroids)

            # Scan the complete original AIRR table
            new_df = pl.concat(
                [
                    pl.scan_csv(
                        f, separator="\t", low_memory=True if large_files else False
                    )
                    for f in files
                ]
            )

            # Filter new_df to only include rows with matching sequence_ids
            filtered_df = new_df.join(
                centroids_lazy.select(["sequence_id"]), on="sequence_id", how="semi"
            ).select(new_df.columns)

            # Add the 'count' column to the filtered DataFrame (only for matching rows)
            if keep_cluster_sizes:
                filtered_df = (
                    filtered_df.join(
                        centroids_lazy.select(["sequence_id", "count"]),
                        on="sequence_id",
                        how="left",
                    )
                    .with_columns(
                        pl.col("count").fill_null(
                            0
                        )  # Replace null values with 0 if necessary
                    )
                    .rename(
                        {
                            "count": "umi_count" if umi else "cluster_size"
                        }  # Rename the column
                    )
                )

            # Sink the resulting DataFrame to a TSV file
            filtered_df.collect().write_csv(tsv_file, separator="\t")
            print(f"Output written to TSV file: {tsv_file}\n")

    end = time.time()
    duration = end - start
    print(
        f"Data reduction complete. Total elapsed time: {datetime.timedelta(seconds=int(duration))}."
    )

    return



import pytest

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

import seaborn as sns

from ..utils.color import (
    get_cmap,
    truncate_colormap,
    hex_to_rgb,
    rgb_to_hex,
    hls,
    husl,
    palettes,
    show_palettes,
)


# ---------------------------
#        colormaps
# ---------------------------


def test_get_cmap_with_colormap():
    cmap = plt.get_cmap("viridis")
    got_cmap = get_cmap(cmap)
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_colormap_name():
    cmap = plt.get_cmap("viridis")
    got_cmap = get_cmap("viridis")
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_hex_code():
    cmap = sns.light_palette("#FF0000", as_cmap=True)
    got_cmap = get_cmap("#FF0000")
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_matplotlib_color():
    cmap = sns.light_palette("#0000FF", as_cmap=True)
    got_cmap = get_cmap("blue")
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_rgb_tuple():
    cmap = sns.light_palette("#00FF00", as_cmap=True)
    got_cmap = get_cmap((0, 1, 0))
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_single_color():
    cmap = sns.light_palette("#FF0000", as_cmap=True)
    got_cmap = get_cmap("#FF0000")
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_dark():
    cmap = sns.dark_palette("#FF0000", as_cmap=True)
    got_cmap = get_cmap("#FF0000", dark=True)
    for i in np.arange(0, 1, 0.1):
        assert np.allclose(got_cmap(i), cmap(i))


def test_get_cmap_with_zero_color():
    got_cmap = get_cmap("#FF0000", zero_color=(0, 0, 0))
    assert got_cmap(0) == (0, 0, 0, 1)


def test_get_cmap_with_minval_and_maxval():
    cmap = get_cmap("#FF0000", name="my_cmap")
    got_cmap = get_cmap("#FF0000", minval=0.25, maxval=0.75, name="truncated_cmap")
    assert np.allclose(got_cmap(0), cmap(0.25))
    assert np.allclose(got_cmap.reversed()(0), cmap.reversed()(0.25), rtol=0.05)
    # assert np.allclose(got_cmap(1), cmap(0.75))


def test_get_cmap_with_name():
    cmap = get_cmap("#FF0000", name="my_cmap")
    cmap.name = "my_cmap"


@pytest.mark.xfail
def test_get_cmap_with_invalid_colormap():
    with pytest.raises(ValueError):
        get_cmap("invalid_colormap")


def test_truncate_colormap_with_full_range():
    cmap = plt.get_cmap("viridis")
    truncated_cmap = truncate_colormap(cmap)
    assert truncated_cmap.name == "viridis-trunc-0-1"
    assert truncated_cmap.N == 256
    assert np.allclose(truncated_cmap(0), cmap(0))
    assert np.allclose(truncated_cmap(1), cmap(1))


def test_truncate_colormap_with_partial_range():
    cmap = plt.get_cmap("viridis")
    truncated_cmap = truncate_colormap(cmap, minval=0.25, maxval=0.75)
    assert truncated_cmap.name == "viridis-trunc-0.25-0.75"
    assert truncated_cmap.N == 256
    assert np.allclose(truncated_cmap(0), cmap(0.25))
    assert np.allclose(truncated_cmap.reversed()(0), cmap(0.75), rtol=0.05)


def test_truncate_colormap_with_reversed_range():
    cmap = plt.get_cmap("viridis")
    truncated_cmap = truncate_colormap(cmap, minval=0.75, maxval=0.25)
    assert truncated_cmap.name == "viridis-trunc-0.75-0.25"
    assert truncated_cmap.N == 256
    assert np.allclose(truncated_cmap(0), cmap(0.75))
    assert np.allclose(truncated_cmap.reversed()(1), cmap(0.25), rtol=0.05)


def test_truncate_colormap_with_custom_n():
    cmap = plt.get_cmap("viridis")
    truncated_cmap = truncate_colormap(cmap, n=10)
    assert truncated_cmap.name == "viridis-trunc-0-1"
    assert truncated_cmap.N == 10


# ---------------------------
#        HEX and RGB
# ---------------------------


def test_hex_to_rgb():
    assert hex_to_rgb("#FFFFFF") == (255, 255, 255)
    assert hex_to_rgb("#000000") == (0, 0, 0)
    assert hex_to_rgb("#FF0000") == (255, 0, 0)
    assert hex_to_rgb("#00FF00") == (0, 255, 0)
    assert hex_to_rgb("#0000FF") == (0, 0, 255)


def test_rgb_to_hex():
    assert rgb_to_hex((255, 255, 255)) == "#ffffff"
    assert rgb_to_hex((0, 0, 0)) == "#000000"
    assert rgb_to_hex((255, 0, 0)) == "#ff0000"
    assert rgb_to_hex((0, 255, 0)) == "#00ff00"
    assert rgb_to_hex((0, 0, 255)) == "#0000ff"


# ---------------------------
#         palettes
# ---------------------------


def test_hls():
    assert hls(3) == sns.hls_palette(3, h=0.01, l=0.6, s=0.65)
    assert hls(5, hue=0.5, lightness=0.4, saturation=0.8) == sns.hls_palette(
        5, h=0.5, l=0.4, s=0.8
    )


def test_husl():
    assert husl(3) == sns.husl_palette(3, h=0.01, s=0.9, l=0.65)
    assert husl(5, hue=0.5, saturation=0.8, lightness=0.4) == sns.husl_palette(
        5, h=0.5, s=0.8, l=0.4
    )


@pytest.mark.skip(reason="Not sure how to test plotting functions")
def test_show_palettes():
    with plt.show._enter_session():
        show_palettes(palettes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import shutil
import tempfile
from pathlib import Path
from unittest import mock

import polars as pl
import pytest

from abutils import Sequence
from abutils.tools import preprocessing


# Fixtures
@pytest.fixture
def temp_dir():
    """Create a temporary directory for test files."""
    dir_path = tempfile.mkdtemp(prefix="tempdir")
    yield dir_path
    shutil.rmtree(dir_path)


@pytest.fixture
def fastq_file(temp_dir):
    """Create a simple FASTQ file for testing."""
    fastq_path = os.path.join(temp_dir, "test_file.fastq")
    with open(fastq_path, "w") as f:
        f.write("test content")
    return fastq_path


@pytest.fixture
def illumina_fastq(temp_dir):
    """Create a FASTQ file with Illumina naming convention."""
    fastq_path = os.path.join(temp_dir, "Sample1_S1_L001_R1_001.fastq")
    with open(fastq_path, "w") as f:
        f.write("@header1\nACGT\n+\nIIII\n")
    return fastq_path


@pytest.fixture
def element_fastq(temp_dir):
    """Create a FASTQ file with Element naming convention."""
    fastq_path = os.path.join(temp_dir, "Sample1_R1.fastq")
    with open(fastq_path, "w") as f:
        f.write("@header1\nACGT\n+\nIIII\n")
    return fastq_path


@pytest.fixture
def illumina_pair(temp_dir):
    """Create a pair of Illumina FASTQ files."""
    r1_path = os.path.join(temp_dir, "Sample1_S1_L001_R1_001.fastq")
    r2_path = os.path.join(temp_dir, "Sample1_S1_L001_R2_001.fastq")

    with open(r1_path, "w") as f:
        f.write("@header1\nACGT\n+\nIIII\n")

    with open(r2_path, "w") as f:
        f.write("@header1\nTGCA\n+\nIIII\n")

    return r1_path, r2_path


@pytest.fixture
def merged_dir(temp_dir):
    """Create an output directory for merged files."""
    merged_dir = os.path.join(temp_dir, "merged")
    os.makedirs(merged_dir, exist_ok=True)
    return merged_dir


@pytest.fixture
def log_dir(temp_dir):
    """Create a directory for log files."""
    log_dir = os.path.join(temp_dir, "logs")
    os.makedirs(log_dir, exist_ok=True)
    return log_dir


@pytest.fixture
def multi_sample_fastqs(temp_dir):
    """Create multiple sample FASTQ files."""
    r1_path = os.path.join(temp_dir, "Sample1_S1_L001_R1_001.fastq")
    r2_path = os.path.join(temp_dir, "Sample1_S1_L001_R2_001.fastq")
    r1_path2 = os.path.join(temp_dir, "Sample2_S2_L001_R1_001.fastq")
    r2_path2 = os.path.join(temp_dir, "Sample2_S2_L001_R2_001.fastq")

    for path in [r1_path, r2_path, r1_path2, r2_path2]:
        with open(path, "w") as f:
            f.write("@header1\nACGT\n+\nIIII\n")

    return {
        "dirname": temp_dir,
        "r1_path": r1_path,
        "r2_path": r2_path,
        "r1_path2": r1_path2,
        "r2_path2": r2_path2,
    }


@pytest.fixture
def airr_files(temp_dir):
    """Create temporary AIRR files for testing."""
    sample1_path = os.path.join(temp_dir, "sample1.tsv")
    sample2_path = os.path.join(temp_dir, "sample2.tsv")

    # Create a simple AIRR dataframe
    data1 = {
        "sequence_id": ["seq1", "seq2", "seq3"],
        "sequence": ["ACGT", "ACGT", "TGCA"],
        "umi": ["UMI1", "UMI2", "UMI3"],
        "v_gene": ["IGHV1-1*01", "IGHV1-2*01", "IGHV2-1*01"],
        "j_gene": ["IGHJ1*01", "IGHJ2*01", "IGHJ1*01"],
        "locus": ["IGH", "IGH", "IGH"],
    }

    data2 = {
        "sequence_id": ["seq4", "seq5", "seq6"],
        "sequence": ["ACGT", "GTCA", "TGCA"],
        "umi": ["UMI4", "UMI5", "UMI6"],
        "v_gene": ["IGHV1-1*01", "IGHV1-2*01", "IGHV2-1*01"],
        "j_gene": ["IGHJ1*01", "IGHJ2*01", "IGHJ1*01"],
        "locus": ["IGH", "IGH", "IGH"],
    }

    df1 = pl.DataFrame(data1)
    df2 = pl.DataFrame(data2)

    df1.write_csv(sample1_path, separator="\t")
    df2.write_csv(sample2_path, separator="\t")

    # Create output directory
    output_dir = os.path.join(temp_dir, "output")
    os.makedirs(output_dir, exist_ok=True)

    return {
        "dirname": temp_dir,
        "sample1_path": sample1_path,
        "sample2_path": sample2_path,
        "output_dir": output_dir,
    }


# Tests for FASTQFile
def test_fastq_file_init(fastq_file):
    """Test initialization of FASTQFile."""
    fastq_obj = preprocessing.FASTQFile(fastq_file)
    assert fastq_obj.file == fastq_file
    assert fastq_obj.path == os.path.abspath(fastq_file)
    assert fastq_obj.basename == "test_file.fastq"
    assert fastq_obj.dir == os.path.dirname(os.path.abspath(fastq_file))
    assert fastq_obj.filename == "test_file"


# Tests for IlluminaFile
def test_illumina_file_init(illumina_fastq):
    """Test initialization of IlluminaFile."""
    illumina_file = preprocessing.IlluminaFile(illumina_fastq)
    assert illumina_file.schema == "illumina"
    assert illumina_file.filename == "Sample1_S1_L001_R1_001"


def test_illumina_file_properties(illumina_fastq):
    """Test properties of IlluminaFile."""
    illumina_file = preprocessing.IlluminaFile(illumina_fastq)
    assert illumina_file.name == "Sample1"
    assert illumina_file.sample == "S1"
    assert illumina_file.lane == "L001"
    assert illumina_file.read == "R1"
    assert illumina_file.number == "001"


# Tests for ElementFile
def test_element_file_init(element_fastq):
    """Test initialization of ElementFile."""
    element_file = preprocessing.ElementFile(element_fastq)
    assert element_file.schema == "element"
    assert element_file.filename == "Sample1_R1"


def test_element_file_properties(element_fastq):
    """Test properties of ElementFile."""
    element_file = preprocessing.ElementFile(element_fastq)
    assert element_file.name == "Sample1"
    assert element_file.read == "R1"
    assert element_file.lane == ""  # Element files don't have lane info
    assert element_file.sample == ""  # Element files don't have sample info
    assert element_file.number == ""  # Element files don't have number info


# Tests for MergeGroup
def test_merge_group_init(illumina_pair):
    """Test initialization of MergeGroup."""
    r1_path, r2_path = illumina_pair
    r1_file = preprocessing.IlluminaFile(r1_path)
    r2_file = preprocessing.IlluminaFile(r2_path)

    merge_group = preprocessing.MergeGroup("Sample1", [r1_file, r2_file])
    assert merge_group.name == "Sample1"
    assert len(merge_group.files) == 2
    assert merge_group.merged_file is None


def test_merge_group_by_lane(illumina_pair):
    """Test _group_by_lane method of MergeGroup."""
    r1_path, r2_path = illumina_pair
    r1_file = preprocessing.IlluminaFile(r1_path)
    r2_file = preprocessing.IlluminaFile(r2_path)

    merge_group = preprocessing.MergeGroup("Sample1", [r1_file, r2_file])
    groups = merge_group._group_by_lane()

    # Should have one group for L001
    assert len(groups) == 1

    # Group should contain both R1 and R2
    assert len(groups[0]) == 2

    # Files should be in the group
    assert r1_file in groups[0]
    assert r2_file in groups[0]


def test_merge_group_merge(illumina_pair, merged_dir, log_dir):
    """Test merge method of MergeGroup."""
    r1_path, r2_path = illumina_pair
    r1_file = preprocessing.IlluminaFile(r1_path)
    r2_file = preprocessing.IlluminaFile(r2_path)

    merge_group = preprocessing.MergeGroup("Sample1", [r1_file, r2_file])
    result = merge_group.merge(
        merged_directory=merged_dir,
        log_directory=log_dir,
        format="fastq",
        algo="fastp",
        verbose=True,
    )

    # Check result
    expected_path = os.path.join(merged_dir, "Sample1.fastq")
    assert result == expected_path
    assert merge_group.merged_file == expected_path


# Tests for group_fastq_pairs
def test_group_fastq_pairs(multi_sample_fastqs):
    """Test grouping of FASTQ pairs."""
    r1_file = preprocessing.IlluminaFile(multi_sample_fastqs["r1_path"])
    r2_file = preprocessing.IlluminaFile(multi_sample_fastqs["r2_path"])
    r1_file2 = preprocessing.IlluminaFile(multi_sample_fastqs["r1_path2"])
    r2_file2 = preprocessing.IlluminaFile(multi_sample_fastqs["r2_path2"])

    files = [r1_file, r2_file, r1_file2, r2_file2]
    merge_groups = preprocessing.group_fastq_pairs(files)

    # Should have two merge groups
    assert len(merge_groups) == 2

    # Check that groups are correctly formed
    sample1_group = next((g for g in merge_groups if g.name == "Sample1"), None)
    sample2_group = next((g for g in merge_groups if g.name == "Sample2"), None)

    assert sample1_group is not None
    assert sample2_group is not None

    assert len(sample1_group.files) == 2
    assert len(sample2_group.files) == 2


def test_merge_fastqs(multi_sample_fastqs, merged_dir, log_dir):
    """Test merge_fastqs function."""

    result = preprocessing.merge_fastqs(
        files=multi_sample_fastqs["dirname"],
        output_directory=merged_dir,
        log_directory=log_dir,
        schema="illumina",
        verbose=True,
    )

    # Should return a list of merged files
    assert len(result) > 0
    # The merged file should be in the output directory
    assert os.path.dirname(result[0]) == merged_dir


def test_merge_fastqs_fastp(illumina_pair, merged_dir, log_dir):
    """Test merge_fastqs_fastp function."""
    r1_path, r2_path = illumina_pair

    merged_file = os.path.join(merged_dir, "merged.fastq")
    result = preprocessing.merge_fastqs_fastp(
        forward=r1_path,
        reverse=r2_path,
        merged=merged_file,
        log_directory=log_dir,
        name="Sample1",
    )

    # Check result
    assert result == merged_file


@mock.patch("abutils.tools.preprocessing.sp.Popen")
def test_merge_fastqs_vsearch(mock_popen, illumina_pair, merged_dir):
    """Test merge_fastqs_vsearch function."""
    r1_path, r2_path = illumina_pair

    # Need to mock subprocess.Popen for external commands
    mock_process = mock.MagicMock()
    mock_process.communicate.return_value = (b"", b"")
    mock_process.returncode = 0
    mock_popen.return_value = mock_process

    merged_file = os.path.join(merged_dir, "merged.fasta")
    result = preprocessing.merge_fastqs_vsearch(
        forward=r1_path,
        reverse=r2_path,
        merged_file=merged_file,
        output_format="fasta",
    )

    # Check result
    assert result == merged_file

    # Check that sp.Popen was called with vsearch command
    cmd = mock_popen.call_args[0][0]
    assert "vsearch" in cmd
    assert "--fastq_mergepairs" in cmd
    assert r1_path in cmd
    assert r2_path in cmd
    assert merged_file in cmd


# Test helper functions
def test_deduplicate_sequences():
    """Test _deduplicate_sequences function."""
    df = pl.DataFrame(
        {"sequence_id": ["seq1", "seq2", "seq3"], "sequence": ["ACGT", "ACGT", "TGCA"]}
    )

    # Test without keeping read numbers
    result = preprocessing._deduplicate_sequences(df, ["sequence"], False)
    assert result.height == 2  # Should have 2 unique sequences

    # Test with keeping read numbers
    result = preprocessing._deduplicate_sequences(df, ["sequence"], True)
    assert result.height == 2  # Should have 2 unique sequences
    assert "count" in result.columns  # Should have a count column
    assert result.select("count").to_series().to_list() == [2, 1]  # Check count values


# # Test for deduplicate and reduction functions
# @mock.patch("abutils.tools.preprocessing.to_fasta")
# def test_deduplicate(mock_to_fasta, airr_files, temp_dir):
#     """Test deduplicate function."""
#     # Mock to_fasta to avoid actual file writing
#     mock_to_fasta.return_value = None

#     output_dir = os.path.join(temp_dir, "dedup")

#     preprocessing.deduplicate(
#         # project_folder=airr_files["sample1_path"],  # Test file path input
#         project_folder=airr_files["dirname"],  # Test file path input
#         output="dedup",
#         output_format="fasta",
#         debug=True,
#     )

#     # Check that to_fasta was called with deduped data
#     assert mock_to_fasta.called
#     # The first argument should be a list of tuples with sequence_id and sequence
#     # We can't assert exact values because of potential mock issues, but we can check structure
#     args = mock_to_fasta.call_args[0]
#     assert isinstance(args[0], list)


@mock.patch("abutils.tools.cluster.cluster")
@mock.patch("abutils.tools.alignment.make_consensus")
def test_process_chains(mock_make_consensus, mock_cluster):
    """Test _process_chains function."""
    # Setup mock cluster
    mock_cluster_obj = mock.MagicMock()
    mock_cluster_obj.size = 3
    mock_cluster_obj.centroid = mock.MagicMock()
    mock_cluster_obj.centroid.id = "seq1"
    mock_cluster_obj.centroid.sequence = "ACGT"
    mock_cluster_obj.sequences = [
        Sequence("ACGT", id="seq1"),
        Sequence("ACGT", id="seq2"),
        Sequence("ACGT", id="seq3"),
    ]

    mock_cluster.return_value = [mock_cluster_obj]

    mock_consensus = Sequence("ACGT", id="consensus")
    mock_make_consensus.return_value = mock_consensus

    # Create test dataframe
    df = pl.DataFrame(
        {
            "sequence_id": ["seq1", "seq2", "seq3"],
            "sequence": ["ACGT", "ACGT", "ACGT"],
            "v_gene": ["IGHV1-1*01", "IGHV1-1*01", "IGHV1-1*01"],
            "j_gene": ["IGHJ1*01", "IGHJ1*01", "IGHJ1*01"],
            "vj_bin": [
                "IGHV1-1*01_IGHJ1*01",
                "IGHV1-1*01_IGHJ1*01",
                "IGHV1-1*01_IGHJ1*01",
            ],
        }
    )

    # Test centroid mode
    result = preprocessing._process_chains(
        "test",
        df,
        min_cluster_size=2,
        clustering_threshold=0.9,
        consentroid="centroid",
        cluster_sizes_separator="|",
        keep_cluster_sizes=True,
        output_format="fasta",
        debug=True,
    )

    assert len(result) == 1
    assert result[0].id == "seq1|3"
    assert result[0].sequence == "ACGT"

    # Test consensus mode
    result = preprocessing._process_chains(
        "test",
        df,
        min_cluster_size=2,
        clustering_threshold=0.9,
        consentroid="consensus",
        cluster_sizes_separator="|",
        keep_cluster_sizes=True,
        output_format="fasta",
        debug=True,
    )

    assert len(result) == 1
    assert result[0].id == "consensus|3"
    assert result[0].sequence == "ACGT"


def test_process_chain_group():
    """Test _process_chain_group function."""
    # Create test dataframe
    df = pl.DataFrame(
        {
            "sequence_id": ["seq1", "seq2", "seq3"],
            "sequence": ["ACGT", "ACGT", "TGCA"],
            "v_gene": ["IGHV1-1*01", "IGHV1-1*01", "IGHV2-1*01"],
            "j_gene": ["IGHJ1*01", "IGHJ1*01", "IGHJ2*01"],
            "locus": ["IGH", "IGH", "IGH"],
        }
    )

    with mock.patch(
        "abutils.tools.preprocessing._process_chains"
    ) as mock_process_chains:
        mock_process_chains.return_value = [Sequence("ACGT", id="seq1")]

        # Test without UMI
        result = preprocessing._process_chain_group(
            df,
            chain_type="Heavy",
            umi=False,
            min_cluster_size=2,
            clustering_threshold=0.9,
            consentroid="centroid",
            cluster_sizes_separator="|",
            keep_cluster_sizes=False,
            output_format="fasta",
            debug=True,
        )

        assert result == [Sequence("ACGT", id="seq1")]

        # Check that vj_bin was created correctly
        assert "vj_bin" in mock_process_chains.call_args[0][1].columns
        vj_bins = (
            mock_process_chains.call_args[0][1].select("vj_bin").to_series().to_list()
        )
        assert "IGHV1-1*01_IGHJ1*01" in vj_bins
        assert "IGHV2-1*01_IGHJ2*01" in vj_bins

        # Test with UMI
        df_with_umi = df.with_columns(pl.lit("UMI1").alias("umi"))

        result = preprocessing._process_chain_group(
            df_with_umi,
            chain_type="Heavy",
            umi=True,
            min_cluster_size=2,
            clustering_threshold=0.9,
            consentroid="centroid",
            cluster_sizes_separator="|",
            keep_cluster_sizes=False,
            output_format="fasta",
            debug=True,
        )

        assert result == [Sequence("ACGT", id="seq1")]

        # Check that vj_bin includes UMI
        vj_bins = (
            mock_process_chains.call_args[0][1].select("vj_bin").to_series().to_list()
        )
        assert any("UMI1" in bin_name for bin_name in vj_bins)


# @mock.patch("abutils.tools.preprocessing.to_fasta")
# @mock.patch("abutils.tools.preprocessing._process_chain_group")
# def test_reduction(mock_process_chain_group, mock_to_fasta, airr_files, temp_dir):
#     """Test reduction function."""
#     mock_process_chain_group.return_value = [
#         Sequence("ACGT", id="seq1"),
#         Sequence("TGCA", id="seq2"),
#     ]

#     mock_to_fasta.return_value = None

#     output_dir = os.path.join(temp_dir, "reduced")

#     preprocessing.reduction(
#         project_folder=airr_files["sample1_path"],  # Test file path input
#         output="reduced",
#         output_format="fasta",
#         debug=True,
#     )

#     # Check that _process_chain_group and to_fasta were called
#     assert mock_process_chain_group.called
#     assert mock_to_fasta.called



import os
import shutil
import tempfile

import pandas as pd
import polars as pl
import pytest

from ..core.pair import Pair
from ..core.sequence import Sequence
from ..io import (
    concatenate_airr,
    concatenate_csv,
    concatenate_parquet,
    from_pandas,
    from_polars,
    split_airr,
    split_csv,
    split_fasta,
    split_fastq,
    split_parquet,
    to_pandas,
    to_polars,
)


# ==================
#      FIXTURES
# ==================


@pytest.fixture
def temp_dir():
    """Create a temporary directory for test outputs."""
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d, ignore_errors=True)


@pytest.fixture
def parquet_file():
    """Create a temp parquet file with 100 rows."""
    data = [
        {"sequence_id": f"seq{i}", "sequence": "ATCGATCG" * 5, "v_gene": f"IGHV{i % 5}"}
        for i in range(100)
    ]
    df = pl.DataFrame(data)
    pf = tempfile.NamedTemporaryFile(delete=False, suffix=".parquet")
    df.write_parquet(pf.name)
    pf.close()
    yield pf.name
    os.unlink(pf.name)


@pytest.fixture
def csv_file():
    """Create a temp CSV file with 100 rows."""
    content = "sequence_id,sequence,v_gene\n"
    content += "\n".join(
        [f"seq{i},ATCGATCGATCG,IGHV{i % 5}" for i in range(100)]
    )
    cf = tempfile.NamedTemporaryFile(delete=False, suffix=".csv", mode="w")
    cf.write(content)
    cf.close()
    yield cf.name
    os.unlink(cf.name)


@pytest.fixture
def airr_file():
    """Create a temp AIRR file (TSV) with 100 rows."""
    content = "sequence_id\tsequence\tv_gene\n"
    content += "\n".join(
        [f"seq{i}\tATCGATCGATCG\tIGHV{i % 5}" for i in range(100)]
    )
    af = tempfile.NamedTemporaryFile(delete=False, suffix=".tsv", mode="w")
    af.write(content)
    af.close()
    yield af.name
    os.unlink(af.name)


@pytest.fixture
def fasta_file():
    """Create a temp FASTA file with 100 sequences."""
    content = "\n".join([f">seq{i}\nATCGATCGATCGATCG" for i in range(100)])
    ff = tempfile.NamedTemporaryFile(delete=False, suffix=".fasta", mode="w")
    ff.write(content)
    ff.close()
    yield ff.name
    os.unlink(ff.name)


@pytest.fixture
def fastq_file():
    """Create a temp FASTQ file with 100 sequences."""
    content = "\n".join(
        [f"@seq{i}\nATCGATCGATCGATCG\n+\nIIIIIIIIIIIIIIII" for i in range(100)]
    )
    fq = tempfile.NamedTemporaryFile(delete=False, suffix=".fastq", mode="w")
    fq.write(content)
    fq.close()
    yield fq.name
    os.unlink(fq.name)


@pytest.fixture
def sequence_list():
    """List of Sequence objects for DataFrame conversion testing."""
    return [
        Sequence(
            {
                "sequence_id": f"seq{i}",
                "sequence": "ATCGATCG",
                "v_gene": f"IGHV{i}",
                "locus": "IGH",
            }
        )
        for i in range(5)
    ]


@pytest.fixture
def pair_list():
    """List of Pair objects for DataFrame conversion testing."""
    pairs = []
    for i in range(5):
        heavy = Sequence(
            {
                "sequence_id": f"pair{i}",
                "sequence": "ATCGATCGATCG",
                "locus": "IGH",
                "v_gene": f"IGHV{i}-1",
            }
        )
        light = Sequence(
            {
                "sequence_id": f"pair{i}",
                "sequence": "GCTAGCTAGCTA",
                "locus": "IGK",
                "v_gene": f"IGKV{i}-1",
            }
        )
        pairs.append(Pair([heavy, light]))
    return pairs


# ============================
#    SPLIT FUNCTION TESTS
# ============================


def test_split_parquet_creates_files(parquet_file, temp_dir):
    """Test split_parquet creates multiple output files."""
    output_files = split_parquet(parquet_file, temp_dir, num_rows=30)
    assert len(output_files) == 4  # 100 rows / 30 = 4 files (with remainder)
    for f in output_files:
        assert os.path.exists(f)
        assert f.endswith(".parquet")


def test_split_parquet_correct_row_count(parquet_file, temp_dir):
    """Test split_parquet produces files with correct row counts."""
    output_files = split_parquet(parquet_file, temp_dir, num_rows=25)
    total_rows = 0
    for f in output_files:
        df = pl.read_parquet(f)
        total_rows += len(df)
    assert total_rows == 100


def test_split_parquet_with_num_splits(parquet_file, temp_dir):
    """Test split_parquet with num_splits parameter."""
    output_files = split_parquet(parquet_file, temp_dir, num_splits=5)
    assert len(output_files) == 5


def test_split_csv_creates_files(csv_file, temp_dir):
    """Test split_csv creates multiple output files."""
    output_files = split_csv(csv_file, temp_dir, chunksize=30)
    assert len(output_files) == 4  # 100 rows / 30 = 4 files
    for f in output_files:
        assert os.path.exists(f)
        assert f.endswith(".csv")


def test_split_airr_creates_files(airr_file, temp_dir):
    """Test split_airr creates multiple output files."""
    output_files = split_airr(airr_file, temp_dir, chunksize=30)
    assert len(output_files) == 4
    for f in output_files:
        assert os.path.exists(f)


def test_split_fasta_creates_files(fasta_file, temp_dir):
    """Test split_fasta creates multiple output files."""
    output_files = split_fasta(fasta_file, temp_dir, chunksize=30)
    assert len(output_files) == 4
    for f in output_files:
        assert os.path.exists(f)
        assert f.endswith(".fasta")


def test_split_fastq_creates_files(fastq_file, temp_dir):
    """Test split_fastq creates multiple output files."""
    output_files = split_fastq(fastq_file, temp_dir, chunksize=30)
    assert len(output_files) == 4
    for f in output_files:
        assert os.path.exists(f)
        assert f.endswith(".fastq")


# ================================
#    CONCATENATE FUNCTION TESTS
# ================================


def test_concatenate_parquet(parquet_file, temp_dir):
    """Test concatenate_parquet merges files correctly."""
    # First split, then concatenate
    split_files = split_parquet(parquet_file, temp_dir, num_rows=30)
    output_file = os.path.join(temp_dir, "concatenated.parquet")
    result = concatenate_parquet(split_files, output_file)
    assert os.path.exists(result)
    df = pl.read_parquet(result)
    assert len(df) == 100


def test_concatenate_csv(csv_file, temp_dir):
    """Test concatenate_csv merges files correctly."""
    # First split, then concatenate
    split_files = split_csv(csv_file, temp_dir, chunksize=30)
    output_file = os.path.join(temp_dir, "concatenated.csv")
    result = concatenate_csv(split_files, output_file)
    assert os.path.exists(result)
    df = pl.read_csv(result)
    assert len(df) == 100


def test_concatenate_airr(airr_file, temp_dir):
    """Test concatenate_airr merges files correctly."""
    # First split, then concatenate
    split_files = split_airr(airr_file, temp_dir, chunksize=30)
    output_file = os.path.join(temp_dir, "concatenated.tsv")
    result = concatenate_airr(split_files, output_file)
    assert os.path.exists(result)
    df = pl.read_csv(result, separator="\t")
    assert len(df) == 100


# ====================================
#    DATAFRAME CONVERSION TESTS
# ====================================


def test_to_polars_with_sequences(sequence_list):
    """Test to_polars with Sequence objects."""
    df = to_polars(sequence_list)
    assert isinstance(df, pl.DataFrame)
    assert len(df) == 5
    assert "sequence_id" in df.columns
    assert "sequence" in df.columns


def test_from_polars_with_sequences():
    """Test from_polars returns Sequence objects."""
    df = pl.DataFrame(
        {
            "sequence_id": ["seq1", "seq2", "seq3"],
            "sequence": ["ATCG", "GCTA", "TTAA"],
            "v_gene": ["IGHV1", "IGHV2", "IGHV3"],
        }
    )
    sequences = from_polars(df)
    assert len(sequences) == 3
    assert all(isinstance(s, Sequence) for s in sequences)
    assert sequences[0]["sequence_id"] == "seq1"


def test_to_pandas_with_sequences(sequence_list):
    """Test to_pandas with Sequence objects."""
    df = to_pandas(sequence_list)
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 5
    assert "sequence_id" in df.columns
    assert "sequence" in df.columns


def test_from_pandas_with_sequences():
    """Test from_pandas returns Sequence objects."""
    df = pd.DataFrame(
        {
            "sequence_id": ["seq1", "seq2", "seq3"],
            "sequence": ["ATCG", "GCTA", "TTAA"],
            "v_gene": ["IGHV1", "IGHV2", "IGHV3"],
        }
    )
    sequences = from_pandas(df)
    assert len(sequences) == 3
    assert all(isinstance(s, Sequence) for s in sequences)


def test_to_polars_with_pairs(pair_list):
    """Test to_polars with Pair objects."""
    df = to_polars(pair_list)
    assert isinstance(df, pl.DataFrame)
    assert len(df) == 5
    # Pair data uses :0 and :1 suffixes
    assert "sequence:0" in df.columns or "sequence_id:0" in df.columns


def test_from_polars_with_pairs():
    """Test from_polars returns Pair objects when data has :0/:1 columns."""
    df = pl.DataFrame(
        {
            "sequence_id:0": ["pair1", "pair2"],
            "sequence:0": ["ATCG", "GCTA"],
            "locus:0": ["IGH", "IGH"],
            "sequence_id:1": ["pair1", "pair2"],
            "sequence:1": ["TTAA", "AACC"],
            "locus:1": ["IGK", "IGK"],
        }
    )
    pairs = from_polars(df)
    assert len(pairs) == 2
    assert all(isinstance(p, Pair) for p in pairs)


def test_to_pandas_with_pairs(pair_list):
    """Test to_pandas with Pair objects."""
    df = to_pandas(pair_list)
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 5


def test_from_polars_with_lazyframe():
    """Test from_polars works with LazyFrame."""
    df = pl.DataFrame(
        {
            "sequence_id": ["seq1", "seq2"],
            "sequence": ["ATCG", "GCTA"],
        }
    ).lazy()
    sequences = from_polars(df)
    assert len(sequences) == 2
    assert all(isinstance(s, Sequence) for s in sequences)


def test_to_polars_with_annotations(sequence_list):
    """Test to_polars with specific annotations."""
    df = to_polars(sequence_list, annotations=["sequence_id", "v_gene"])
    assert "sequence_id" in df.columns
    assert "v_gene" in df.columns


def test_to_polars_drop_na_columns(sequence_list):
    """Test to_polars drop_na_columns parameter."""
    # Add a sequence with extra annotation
    sequence_list[0].annotations["extra_field"] = "value"
    df = to_polars(sequence_list, drop_na_columns=True)
    # extra_field should be dropped since most sequences don't have it
    # (or kept if implementation doesn't drop)
    assert isinstance(df, pl.DataFrame)



#!/usr/bin/env python
# filename: pipeline.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import glob
import os
import re
from typing import Iterable, Optional, Union

from natsort import natsorted

# =======================
#   General file I/O
# =======================


def make_dir(directory: str) -> None:
    """
    Makes a directory, if it doesn't already exist.

    Parameters
    ----------
    directory : str
        Path to a directory.

    """
    if not os.path.exists(directory):
        os.makedirs(os.path.abspath(directory))


def list_files(
    directory: str,
    extension: Union[str, Iterable, None] = None,
    recursive: bool = False,
    match: Optional[str] = None,
    ignore_dot_files: bool = True,
) -> Iterable[str]:
    """
    Lists files in a given directory.

    Parameters
    ----------
    directory : str
        Path to a directory. If a file path is passed instead, the returned list of files will contain
        only that file path.

    extension : str
        If supplied, only files that end with the specificied extension(s) will be returned. Can be either
        a string or a list of strings. Extension evaluation is case-insensitive and can match complex
        extensions (e.g. '.fastq.gz'). Default is ``None``, which returns all files in the directory,
        regardless of extension.

    recursive : bool, default=False
        If ``True``, the directory will be searched recursively, and all files in all subdirectories will be returned.

    match : str, optional
        If supplied, only files that match the specified pattern will be returned. Regular expressions are supported.

    ignore_dot_files : bool, default=True
        If ``True``, dot files (hidden files) will be ignored.

    Returns
    -------
    Iterable[str]

    """
    directory = os.path.abspath(directory)
    if os.path.exists(directory):
        if os.path.isdir(directory):
            if recursive:
                files = []
                for root, dirs, _files in os.walk(directory):
                    for f in _files:
                        file_path = os.path.join(root, f)
                        files.append(file_path)
            else:
                files = natsorted(glob.glob(directory + "/*"))
        else:
            files = [directory]
    else:
        raise ValueError(f"Directory {directory} does not exist.")
    if extension is not None:
        if isinstance(extension, str):
            extension = [
                extension,
            ]
        files = [
            f
            for f in files
            if any(
                [
                    any([f.lower().endswith(e.lower()) for e in extension]),
                    any([f.endswith(e.upper()) for e in extension]),
                    any([f.endswith(e.lower()) for e in extension]),
                ]
            )
        ]
    if ignore_dot_files:
        files = [f for f in files if not os.path.basename(f).startswith(".")]
    if match is not None:
        files = [f for f in files if re.match(match, os.path.basename(f)) is not None]
    return files


def rename_file(file: str, new_name: str) -> None:
    """
    Renames a file.

    Parameters
    ----------
    file : str
        Path to the file to be renamed.

    new_name : str
        New name for the file.

    """
    os.rename(file, new_name)


def delete_files(files: Union[str, Iterable]) -> None:
    """
    Deletes files.

    Parameters
    ----------
    files : Union[str, Iterable]
        Path to a file or an iterable of file paths.

    """
    if isinstance(files, str):
        files = [
            files,
        ]
    for f in files:
        if os.path.exists(f):
            os.remove(f)


# from __future__ import absolute_import, division, print_function, unicode_literals

# import glob
# import os
# import sys
# from typing import Iterable, Optional, Union

# from . import log


# # for backward compatibility
# def list_files(*args, **kwargs):
#     from ..io import list_files

#     return list_files(*args, **kwargs)


# def make_dir(*args, **kwargs):
#     from ..io import make_dir

#     return make_dir(*args, **kwargs)


# def initialize(log_file, project_dir=None, debug=False):
#     """
#     Initializes an AbTools pipeline.

#     Initialization includes printing the AbTools splash, setting up logging,
#     creating the project directory, and logging both the project directory
#     and the log location.

#     Parameters
#     ----------
#     log_file : str
#         Path to the log file. Required.

#     project_dir : str
#         Path to the project directory. If not provided,
#         the project directory won't be created and the location won't be logged.

#     debug : bool
#         If ``True``, the logging level will be set to ``logging.DEBUG``.

#     Returns
#     -------
#     logger
#         A logger instance.

#     """
#     print_splash()
#     log.setup_logging(log_file, print_log_location=False, debug=debug)
#     logger = log.get_logger("pipeline")
#     if project_dir is not None:
#         make_dir(os.path.normpath(project_dir))
#         logger.info("PROJECT DIRECTORY: {}".format(project_dir))
#         logger.info("")
#     logger.info("LOG LOCATION: {}".format(log_file))
#     print("")
#     return logger


# # def make_dir(directory: str) -> None:
# #     """
# #     Makes a directory, if it doesn't already exist.

# #     Parameters
# #     ----------
# #     directory : str
# #         Path to a directory.

# #     """
# #     if not os.path.exists(directory):
# #         os.makedirs(os.path.abspath(directory))


# # def list_files(
# #     directory: str, extension: Union[str, Iterable, None] = None
# # ) -> Iterable[str]:
# #     """
# #     Lists files in a given directory.

# #     Parameters
# #     ----------
# #     directory : str
# #         Path to a directory.

# #     extension : str
# #         If supplied, only files that end with the specificied extension(s) will be returned. Can be either
# #         a string or a list of strings. Extension evaluation is case-insensitive and can match complex
# #         extensions (e.g. '.fastq.gz'). Default is ``None``, which returns all files in the directory,
# #         regardless of extension.

# #     Returns
# #     -------
# #     Iterable[str]

# #     """
# #     if os.path.isdir(directory):
# #         expanded_dir = os.path.expanduser(directory)
# #         files = sorted(glob.glob(expanded_dir + "/*"))
# #     else:
# #         files = [
# #             directory,
# #         ]
# #     if extension is not None:
# #         if isinstance(extension, str):
# #             extension = [
# #                 extension,
# #             ]
# #         files = [
# #             f
# #             for f in files
# #             if any(
# #                 [
# #                     any([f.lower().endswith(e.lower()) for e in extension]),
# #                     any([f.endswith(e.upper()) for e in extension]),
# #                     any([f.endswith(e.lower()) for e in extension]),
# #                 ]
# #             )
# #         ]
# #     return files


# def print_splash():
#     splash = """
#      _    _     _   _ _   _ _       ____  _            _ _
#     / \  | |__ | | | | |_(_) |___  |  _ \(_)_ __   ___| (_)_ __   ___
#    / _ \ | '_ \| | | | __| | / __| | |_) | | '_ \ / _ \ | | '_ \ / _ \\
#   / ___ \| |_) | |_| | |_| | \__ \ |  __/| | |_) |  __/ | | | | |  __/
#  /_/   \_\_.__/ \___/ \__|_|_|___/ |_|   |_| .__/ \___|_|_|_| |_|\___|
#                                            |_|
#     """
#     print("")
#     print(splash)
#     print(
#         "(c) 2023 Bryan Briney\nDistributed under the MIT License (http://opensource.org/licenses/MIT)"
#     )
#     print("")



#!/usr/bin/env python
# filename: cluster.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# import os
# import platform
# import random
# import shutil
# import string
# import subprocess as sp
# import sys
# import tempfile
# from typing import Iterable, Optional, Union

# from Bio.Align import AlignInfo

# from ..core.sequence import Sequence
# from ..io import read_fasta, to_fasta
# from .alignment import mafft
# from .decorators import lazy_property
# from .pipeline import make_dir


# class Cluster:
#     """
#     Docstring for Cluster.
#     """

#     def __init__(self, name, sequences, centroid=None):
#         self.name = name
#         self.sequences = sequences
#         self.centroid = centroid

#     def __iter__(self):
#         for s in self.sequences:
#             yield s

#     @property
#     def size(self):
#         return len(self.sequences)

#     @property
#     def seq_ids(self):
#         return [s.id for s in self.sequences]

#     @lazy_property
#     def consensus(self):
#         return self._make_consensus()

#     def _make_consensus(self):
#         if len(self.sequences) == 1:
#             return self.sequences[0]
#         aln = mafft(self.sequences)
#         if aln is None:
#             print("ERROR: Failed to generate an alignmnet for a consensus sequence.")
#             return None
#         summary_align = AlignInfo.SummaryInfo(aln)
#         consensus = summary_align.gap_consensus(threshold=0.51, ambiguous="n")
#         consensus_string = str(consensus).replace("-", "")
#         consensus_seq = Sequence(consensus_string.upper())
#         return consensus_seq


# class Clusters:
#     """
#     Docstring for Clusters.
#     """

#     def __init__(self, clusters=None):
#         self._clusters = self._parse_clusters(clusters)

#     def __iter__(self):
#         for cluster in self.clusters:
#             yield cluster

#     def __getitem__(self, key):
#         return self.clusters[key]

#     def __len__(self):
#         return len(self._clusters)

#     @property
#     def clusters(self):
#         return sorted(self._clusters, key=lambda x: x.size, reverse=True)

#     @property
#     def centroids(self):
#         return [c.centroid for c in self.clusters]

#     @property
#     def largest_cluster(self):
#         return self.clusters[0]

#     @property
#     def count(self):
#         return len(self.clusters)

#     def _parse_clusters(self, clusts):
#         if isinstance(clusts, dict):
#             clusters = []
#             for name, cdata in clusts.items():
#                 seqs = cdata["seqs"]
#                 centroid = cdata["centroid"]
#                 clusters.append(Cluster(name, seqs, centroid=centroid))
#             return clusters
#         elif clusts is None:
#             return []
#         else:
#             return clusts

#     def add(self, cluster):
#         self._clusters.append(cluster)


# def cluster(
#     sequences: Union[Iterable, str],
#     threshold: float = 0.975,
#     algo: str = "auto",
#     temp_dir: str = "/tmp",
#     iddef: int = 0,
#     vsearch_bin: str = None,
#     mmseqs_bin: str = None,
#     id_key: Optional[str] = None,
#     seq_key: Optional[str] = None,
#     strand: str = "plus",
#     as_dict: bool = False,
#     debug: bool = False,
# ) -> Union[dict, Clusters]:
#     """
#     Clusters sequences using `VSEARCH`_ or `MMseqs2`_. By default, sequences will
#     be clustered with VSEARCH if there are fewer than 10,000 sequences and
#     with MMseqs2 if there are more than 10,000 sequences. These defaults can be
#     overridden with `algo`.

#     Parameters
#     ----------
#     sequences : iterable or string
#         Input sequences in any of the following formats:
#             1. list of abutils ``Sequence`` objects
#             2. FASTA-formatted string
#             3. path to a FASTA-formatted file
#             4. list of BioPython ``SeqRecord`` objects
#             5. list of lists/tuples, of the format ``[sequence_id, sequence]``
#         Required.

#     threshold : float, default=0.975
#         Identity threshold for clustering. Must be between 0 and 1.

#     algo : float, default="auto"
#         Algorithm to be used for clustering. Options are ``"vsearch"``, ``"mmseqs"``,
#         or ``"auto"``. By default (``"auto"``), VSEARCH will be used if there are fewer than 10,000
#         sequences and MMseqs2 will be used for 10,000 sequences or more. Providing ``"vsearch"``
#         or ``"mmseqs"`` will force the use of the desired clustering algorithm regardless of the
#         number or sequences to be clustered.

#     temp_dir : str, default="/tmp"
#         Path to a directory for temporary storage of clustering files.

#     iddef : int, default=1
#         Identity definition, as implemented by VSEARCH. Options are:
#             0. CD-HIT definition: (matching columns) / (shortest sequence length).
#             1. edit distance: (matching columns) / (alignment length).
#             2. edit distance excluding terminal gaps (same as --id).
#             3. Marine Biological Lab definition counting each gap opening (internal or
#                terminal) as a single mismatch, whether or not the gap was extended: 1.0
#                - [(mismatches + gap openings)/(longest sequence length)]
#             4. BLAST definition, equivalent to --iddef 1 in a context of global pairwise
#                alignment.

#     vsearch_bin : str, optional
#         Path to a VSEARCH executable. If not provided, the VSEARCH binary bundled
#         with ``abutils`` will be used.

#     mmseqs_bin : str, optional
#         Path to a MMseqs2 executable. If not provided, the MMseqs2 binary bundled
#         with ``abutils`` will be used.

#     id_key : str, default=None
#         Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

#     sequence_key : str, default=None
#         Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

#     strand : str, default="plus"
#         Strand of the sequences to align. Options are ``"plus"`` and ``"both"``.

#     as_dict : bool, default=False
#         If ``True``, return clustering results as a ``dict`` rather than a ``Clusters``
#         object. the ``dict`` is of the format:
#             {"cluster1_name": {"centroid": cluster1_centroid,
#                                "seqs": [seq1, seq2, seq3, ...]},
#              "cluster2_name": {"centroid": cluster2_centroid,
#                                "seqs": [seq4, seq5, seq6, ...]},
#             }

#     debug : bool, default=False
#         If ``True``, prints MAFFT's standard output and standard error.
#         Default is ``False``.


#     Returns
#     -------
#     clusters : ``Clusters`` or ``dict``


#     .. _VSEARCH
#         https://github.com/torognes/vsearch
#         x
#     .. _MMseqs2
#         https://github.com/soedinglab/MMseqs2

#     """
#     # check input data to get the number of sequences
#     fasta_file = to_fasta(
#         sequences, tempfile_dir=temp_dir, id_key=id_key, sequence_key=seq_key
#     )
#     seqs = read_fasta(fasta_file)
#     seq_dict = {s.id: s for s in seqs}
#     # select the clustering algo
#     algo = algo.lower()
#     if algo == "auto" and len(seqs) < 10000:
#         algo = "vsearch"
#     elif algo == "auto" and len(seqs) >= 10000:
#         algo = "mmseqs"
#     if algo in ["mmseqs", "mmseqs2"]:
#         cluster_dict = cluster_mmseqs(
#             fasta_file=fasta_file,
#             threshold=threshold,
#             temp_dir=temp_dir,
#             mmseqs_bin=mmseqs_bin,
#             as_dict=True,
#             debug=debug,
#         )
#     elif algo == "vsearch":
#         cluster_dict = cluster_vsearch(
#             fasta_file=fasta_file,
#             threshold=threshold,
#             temp_dir=temp_dir,
#             iddef=iddef,
#             vsearch_bin=vsearch_bin,
#             strand=strand,
#             as_dict=True,
#             debug=debug,
#         )
#     else:
#         err = f"\nERROR: Invalid algo option: {algo}."
#         err += " Valid choices are: 'vsearch', 'mmseqs', or 'auto'.\n"
#         print(err)
#         sys.exit()
#     cluster_info = {}
#     for cname, cdata in cluster_dict.items():
#         cluster_info[cname] = {
#             "centroid": seq_dict[cdata["centroid_id"]],
#             "seqs": [seq_dict[seq_id] for seq_id in cdata["seq_ids"]],
#         }
#     if as_dict:
#         return cluster_info
#     return Clusters(cluster_info)


# def cluster_vsearch(
#     fasta_file: str,
#     threshold: float = 0.975,
#     temp_dir: str = "/tmp",
#     iddef: int = 0,
#     vsearch_bin: str = None,
#     strand: str = "plus",
#     as_dict: bool = False,
#     debug: bool = False,
# ) -> Union[dict, Clusters]:
#     """
#     Clusters sequences using `VSEARCH`_.

#     Parameters
#     ----------
#     fasta_file : string
#         Path to a FASTA-formatted file. Required. If you'd like to run ``vsearch``
#         using ``Sequence`` objects as input, use ``cluster(algo="vsearch")``.

#     threshold : float, default=0.975
#         Identity threshold for clustering. Must be between 0 and 1.

#     temp_dir : str, default="/tmp"
#         Path to a directory for temporary storage of clustering files.

#     iddef : int, default=1
#         Identity definition, as implemented by VSEARCH. Options are:
#             0. CD-HIT definition: (matching columns) / (shortest sequence length).
#             1. edit distance: (matching columns) / (alignment length).
#             2. edit distance excluding terminal gaps (same as --id).
#             3. Marine Biological Lab definition counting each gap opening (internal or
#                terminal) as a single mismatch, whether or not the gap was extended: 1.0
#                - [(mismatches + gap openings)/(longest sequence length)]
#             4. BLAST definition, equivalent to --iddef 1 in a context of global pairwise
#                alignment.

#     vsearch_bin : str, optional
#         Path to a VSEARCH executable. If not provided, the VSEARCH binary bundled
#         with ``abutils`` will be used.

#     id_key : str, default=None
#         Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

#     sequence_key : str, default=None
#         Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

#     strand : str, default="plus"
#         Strand of the sequences to align. Options are ``"plus"`` and ``"both"``.

#     as_dict : bool, default=False
#         If ``True``, return clustering results as a ``dict`` rather than a ``Clusters``
#         object. the ``dict`` is of the format:
#             {"cluster1_name": {"centroid": cluster1_centroid,
#                                "seqs": [seq1, seq2, seq3, ...]},
#              "cluster2_name": {"centroid": cluster2_centroid,
#                                "seqs": [seq4, seq5, seq6, ...]},
#             }

#     debug : bool, default=False
#         If ``True``, prints standard output and standard error from ``vsearch``.
#         Default is ``False``.


#     Returns
#     -------
#     clusters : Path to the UC output file from ``vsearch`` or a ``dict`` of cluster info.


#     .. _VSEARCH
#         https://github.com/torognes/vsearch

#     """
#     # output files
#     centroid_file = tempfile.NamedTemporaryFile(
#         dir=temp_dir, delete=False, prefix="centroids_"
#     ).name
#     uc_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False, prefix="uc_").name
#     # get the vsearch binary
#     if vsearch_bin is None:
#         mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#         vsearch_bin = os.path.join(
#             mod_dir, "bin/vsearch_{}".format(platform.system().lower())
#         )
#     # do clustering
#     vsearch_cmd = f"{vsearch_bin} --cluster_fast {fasta_file}"
#     vsearch_cmd += f" --centroids {centroid_file}"
#     vsearch_cmd += f" --clusterout_id"
#     vsearch_cmd += f" --uc {uc_file}"
#     vsearch_cmd += f" --id {threshold}"
#     vsearch_cmd += f" --iddef {iddef}"
#     vsearch_cmd += f" --sizeout"
#     vsearch_cmd += f" --strand {strand}"
#     p = sp.Popen(vsearch_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         print("STDOUT:", stdout.decode("utf-8"))
#         print("")
#         print("STDERR:", stderr.decode("utf-8"))
#     # process output
#     if os.stat(uc_file).st_size == 0:
#         err = f"WARNING: the VSEARCH output file ({uc_file}) is empty. "
#         err += "Please verify that the input data is valid."
#         print(err)
#         return None
#     if as_dict:
#         cluster_info = {}
#         with open(uc_file, "r") as f:
#             for line in f:
#                 if not (ldata := line.strip().split()):
#                     continue
#                 cluster_num = ldata[1]
#                 if cluster_num not in cluster_info:
#                     cluster_info[cluster_num] = {"seq_ids": []}
#                 if ldata[0] == "S":
#                     centroid = ldata[-2]
#                     cluster_info[cluster_num]["centroid_id"] = centroid
#                     cluster_info[cluster_num]["seq_ids"].append(centroid)
#                 elif ldata[0] == "H":
#                     hit = ldata[-2]
#                     cluster_info[cluster_num]["seq_ids"].append(hit)
#         if not debug:
#             os.remove(uc_file)
#         return cluster_info
#     else:
#         return uc_file


# def cluster_mmseqs(
#     fasta_file: str,
#     threshold: float = 0.975,
#     temp_dir: str = "/tmp",
#     mmseqs_bin: str = None,
#     as_dict: bool = False,
#     debug: bool = False,
# ):
#     """
#     Clusters sequences using `MMseqs2`_.

#     Parameters
#     ----------
#     fasta_file : string
#         Path to a FASTA-formatted file. Required. If you'd like to run ``mmseqs``
#         using ``Sequence`` objects as input, use ``cluster(algo="mmseqs")``.

#     threshold : float, default=0.975
#         Identity threshold for clustering. Must be between 0 and 1.

#     temp_dir : str, default="/tmp"
#         Path to a directory for temporary storage of clustering files.

#     mmseqs_bin : str, optional
#         Path to a MMseqs2 executable. If not provided, the MMseqs2 binary bundled
#         with ``abutils`` will be used.

#     id_key : str, default=None
#         Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

#     sequence_key : str, default=None
#         Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

#     as_dict : bool, default=False
#         If ``True``, return clustering results as a ``dict`` rather than the TSV output file.
#         object. the ``dict`` is of the format:
#             {"cluster1_name": {"centroid_id": cluster1_centroid_id,
#                                "seq_ids": [seq1_id, seq2_id, seq3_id, ...]},
#              "cluster2_name": {"centroid_id": cluster2_centroid_id,
#                                "seq_ids": [seq4_id, seq5_id, seq6_id, ...]},
#             }

#     debug : bool, default=False
#         If ``True``, prints standard output and standard error from ``mmseqs``.
#         Default is ``False``.


#     Returns
#     -------
#     clusters : Path to the TSV output file from ``mmseqs`` or a ``dict`` of cluster info.


#     .. _MMseqs2
#         https://github.com/soedinglab/MMseqs2

#     """
#     # output files
#     db_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False, prefix="DB_").name
#     clu_file = tempfile.NamedTemporaryFile(
#         dir=temp_dir, delete=False, prefix="CLU_"
#     ).name
#     tsv_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False, prefix="TSV_")
#     # get the mmseqs binary
#     if mmseqs_bin is None:
#         mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#         mmseqs_bin = os.path.join(
#             mod_dir, "bin/mmseqs_{}".format(platform.system().lower())
#         )
#     # build the mmseqs DB
#     db_cmd = f"{mmseqs_bin} createdb {fasta_file} {db_file}"
#     p = sp.Popen(db_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)
#     # do the clustering
#     cluster_cmd = f"{mmseqs_bin} cluster"
#     cluster_cmd += f" {db_file} {clu_file} {temp_dir}"
#     cluster_cmd += f" --min-seq-id {threshold}"
#     p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)
#     # generate TSV-formatted output
#     tsv_cmd = f"{mmseqs_bin} createtsv"
#     tsv_cmd += f" {db_file} {db_file} {clu_file} {tsv_file}"
#     p = sp.Popen(tsv_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)
#     # parse TSV output
#     if as_dict:
#         cluster_info = {}
#         name_dict = {}
#         with open(tsv_file) as f:
#             for line in f:
#                 line = line.strip()
#                 if line.startswith("#"):
#                     continue
#                 c, s = line.split()
#                 if c not in name_dict:
#                     name = "".join(
#                         random.choice(string.ascii_uppercase + string.digits)
#                         for _ in range(8)
#                     )
#                     name_dict[c] = name
#                 name = name_dict[c]
#                 if name not in cluster_info:
#                     cluster_info[name] = {"centroid_id": c, "seq_ids": []}
#                 cluster_info[name]["seq_ids"].append(s)
#         if not debug:
#             os.remove(tsv_file)
#         return cluster_info
#     else:
#         return tsv_file


# def vsearch_clusterfast(
#     fasta_file, threshold=0.975, temp_dir="/tmp", debug=False, strand="plus"
# ):
#     # preflight
#     fasta_file = os.path.abspath(fasta_file)
#     temp_dir = os.path.join(os.path.abspath(temp_dir), "vsearch")
#     make_dir(temp_dir)
#     mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#     vsearch_bin = os.path.join(
#         mod_dir, "bin/vsearch_{}".format(platform.system().lower())
#     )

#     # do the clustering
#     centroid_file = os.path.join(temp_dir, "centroids")
#     uc_file = os.path.join(temp_dir, "uc")
#     cluster_cmd = (
#         f"{vsearch_bin} --cluster_fast {fasta_file} --centroids {centroid_file}"
#     )
#     cluster_cmd += f" --clusterout_id --uc {uc_file} --id {threshold}"
#     cluster_cmd += f" --iddef 0 --sizeout --strand {strand}"
#     p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         print("STDOUT:", stdout.decode("utf-8"))
#         print("")
#         print("STDERR:", stderr.decode("utf-8"))

#     # parse vsearch output
#     cluster_info = {}
#     with open(uc_file) as f:
#         for line in f:
#             ldata = line.strip().split()
#             cluster_num = ldata[1]
#             if cluster_num not in cluster_info:
#                 cluster_info[cluster_num] = {"seq_ids": []}
#             if ldata[0] == "S":
#                 centroid = ldata[-2]
#                 cluster_info[cluster_num]["centroid"] = centroid
#                 cluster_info[cluster_num]["seq_ids"].append(centroid)
#             elif ldata[0] == "H":
#                 hit = ldata[-2]
#                 cluster_info[cluster_num]["seq_ids"].append(hit)

#     # clean up
#     if not debug:
#         shutil.rmtree(temp_dir)
#     return cluster_info


# def mmseqs_cluster(fasta_file, min_seq_id=0.975, temp_dir="/tmp", debug=False):
#     # preflight
#     fasta_file = os.path.abspath(fasta_file)
#     temp_dir = os.path.join(temp_dir, "mmseqs")
#     make_dir(temp_dir)
#     mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
#     mmseqs_bin = os.path.join(
#         mod_dir, "bin/mmseqs_{}".format(platform.system().lower())
#     )

#     # create the sequence DB
#     db_file = os.path.join(temp_dir, "DB")
#     db_cmd = "{} createdb {} {}".format(mmseqs_bin, fasta_file, db_file)
#     p = sp.Popen(db_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         if sys.version_info[0] > 2:
#             stdout = stdout.decode("utf-8")
#             stderr = stderr.decode("utf-8")
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)

#     # do the clustering
#     clu_file = os.path.join(temp_dir, "CLU")
#     cluster_cmd = "{} cluster {} {} {} --min-seq-id {}".format(
#         mmseqs_bin, db_file, clu_file, temp_dir, min_seq_id
#     )
#     p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         if sys.version_info[0] > 2:
#             stdout = stdout.decode("utf-8")
#             stderr = stderr.decode("utf-8")
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)

#     # generate TSV-formatted output
#     tsv_file = os.path.join(temp_dir, "CLU.tsv")
#     tsv_cmd = "{0} createtsv {1} {1} {2} {3}".format(
#         mmseqs_bin, db_file, clu_file, tsv_file
#     )
#     p = sp.Popen(tsv_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = p.communicate()
#     if debug:
#         if sys.version_info[0] > 2:
#             stdout = stdout.decode("utf-8")
#             stderr = stderr.decode("utf-8")
#         print("STDOUT:", stdout)
#         print("")
#         print("STDERR:", stderr)

#     # parse TSV output
#     clu_names = {}
#     clu_seqs = {}
#     with open(tsv_file) as f:
#         for line in f:
#             line = line.strip()
#             if line.startswith("#"):
#                 continue
#             c, s = line.split()
#             if c not in clu_names:
#                 cname = "".join(
#                     random.choice(string.ascii_uppercase + string.digits)
#                     for _ in range(8)
#                 )
#                 clu_names[c] = cname
#             clu_name = clu_names[c]
#             if clu_name not in clu_seqs:
#                 clu_seqs[clu_name] = {"centroid": c, "seq_ids": []}
#             clu_seqs[clu_name]["seq_ids"].append(s)

#     # clean up
#     if not debug:
#         shutil.rmtree(temp_dir)
#     return clu_seqs


# __all__ = ['CDHITResult', 'Cluster', 'cluster', 'cdhit', 'parse_clusters']


# class CDHITResult(object):
#     '''
#     Docstring for CDHITResult.
#     '''
#     def __init__(self, clusters, seq_db=None, db_path=None, seq_dict=None):
#         self.clusters = sorted(clusters, key = lambda x: x.size, reverse=True)
#         self.seq_db = seq_db
#         self.db_path = db_path
#         self.seq_dict = seq_dict

#     def __len__(self):
#         return len(self.clusters)

#     def __iter__(self):
#         for cluster in self.clusters:
#             yield cluster


#     @property
#     def largest_cluster(self):
#         return self.clusters[0]


#     def delete(self):
#         self.seq_db.connection.close()
#         os.unlink(self.db_path)


# class Cluster(object):
#     """
#     Data and methods for a cluster of sequences.

#     All public attributes are evaluated lazily, so attributes that
#     require significant processing time are only computed when needed.
#     In addition, attributes are cached after calculation, so if you
#     change the Cluster object after accessing attributes, the
#     attributes will not automatically update. Setters are provided for all attributes,
#     however, so you can update them manually if necessary::

#         seqs = [Sequence1, Sequence2, ... SequenceN]
#         clust = cluster(seqs)

#         # calculate the consensus
#         consensus = clust.consensus

#         # add sequences to the Cluster
#         more_sequences = [SequenceA, SequenceB, SequenceC]
#         clust.sequences += more_sequences

#         # need to recompute the consensus manually
#         clust.consensus = clust._make_consensus()


#     Attributes:

#         ids (list): A list of all sequence IDs in the Cluster

#         size (int): Number of sequences in the Cluster

#         sequences (list): A list of all sequences in the Cluster,
#             as AbTools ``Sequence`` objects.

#         consensus (Sequence): Consensus sequence, calculated by
#             aligning all sequences with MAFFT and computing the
#             ``Bio.Align.AlignInfo.SummaryInfo.gap_consensus()``

#         centroid (Sequence): Centroid sequence, as calculated by
#             CD-HIT.
#     """
#     def __init__(self, raw_cluster, seq_db=None, db_path=None, seq_dict=None):
#         super(Cluster, self).__init__()
#         self._raw_cluster = raw_cluster
#         self._seq_db = seq_db
#         self._seq_db_path = db_path
#         self._seq_dict = seq_dict


#     @lazy_property
#     def ids(self):
#         return self._get_ids()

#     @lazy_property
#     def size(self):
#         return len(self.ids)

#     @lazy_property
#     def sequences(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to access a Cluster's sequences, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._get_sequences()

#     @lazy_property
#     def consensus(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to compute a Cluster's consensus, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._make_consensus()

#     @lazy_property
#     def centroid(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to retrieve a Cluster's centroid, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._get_centroid()


#     def cleanup(self):
#         try:
#             os.unlink(self._raw_cluster)
#         except:
#             pass
#         self.terminate_db()


#     def terminate_db(self):
#         if self._seq_db is not None:
#             self._seq_db.close()
#         if self._seq_db_path is not None:
#             if os.path.isfile(self._seq_db_path):
#                 os.unlink(self._seq_db_path)


#     def _get_ids(self):
#         ids = []
#         for c in self._raw_cluster[1:]:
#             if c:
#                 ids.append(c.split()[2][1:-3])
#         return ids

#     def _get_sequences(self):
#         if self._seq_db is not None:
#             seqs = []
#             for chunk in self._chunker(self.ids):
#                 sql_cmd = '''SELECT seqs.id, seqs.sequence
#                              FROM seqs
#                              WHERE seqs.id IN ({})'''.format(','.join('?' * len(chunk)))
#                 seq_chunk = self._seq_db.execute(sql_cmd, chunk)
#                 seqs.extend(seq_chunk)
#             return [Sequence(s) for s in seqs]
#         else:
#             return [self._seq_dict[s] for s in self.ids]

#     def _get_centroid(self):
#         for line in self._raw_cluster:
#            if '*' in line:
#                 centroid_id = line.split()[2][1:-3]
#                 break
#         centroid = [s for s in self.sequences if s.id == centroid_id][0]
#         return centroid

#     def _make_consensus(self):
#         if len(self.sequences) == 1:
#             return self.sequences[0]
#         _aln = mafft(self.sequences, as_file=True)
#         aln = AlignIO.read(open(_aln, 'r'), 'fasta')
#         summary_align = AlignInfo.SummaryInfo(aln)
#         consensus = summary_align.gap_consensus(threshold=0.51, ambiguous='n')
#         consensus_string = str(consensus).replace('-', '')
#         consensus_seq = Sequence(consensus_string.upper())
#         os.unlink(_aln)
#         return consensus_seq

#     @staticmethod
#     def _chunker(l, size=900):
#         return (l[pos:pos + size] for pos in range(0, len(l), size))


# def cluster(seqs, threshold=0.975, out_file=None, temp_dir=None, make_db=True, method='cdhit',
#             quiet=False, threads=0, return_just_seq_ids=False, max_memory=800, retries=5, debug=False):
#     '''
#     Perform sequence clustering with CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that ``abutils.utils.sequence.Sequence``
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is ``0.975``.

#         out_file (str): Path to the clustering output file. Default is to use
#             ``tempfile.NamedTemporaryFile`` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, ``'/tmp'`` is used.

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs)
#             Default is ``True``.

#         method (str): Doesn't do anything at this point, stubbed to allow expansion of sequencing tools
#             beyond CD-HIT (UCLUST, etc).

#         quiet (bool): If ``True``, surpresses printing of output/progress info. Default is ``False``.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is ``0``,
#             which results in all available cores being used.

#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the ``-M`` runtime option. Default is ``800``.

#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty.
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are
#             not empty. Default is 5.

#         debug (bool): If ``True``, print standard output and standard error from CD-HIT. Default is ``False``.

#     Returns:

#         list: A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     if make_db:
#         ofile, cfile, seq_db, db_path = cdhit(seqs, out_file=out_file, temp_dir=temp_dir,
#                                               threshold=threshold, make_db=True, quiet=quiet,
#                                               threads=threads, max_memory=max_memory, debug=debug)
#         return parse_clusters(ofile, cfile, seq_db=seq_db, db_path=db_path, return_just_seq_ids=return_just_seq_ids)
#     else:
#         if isinstance(seqs, (list, tuple)):
#             seqs = [Sequence(s) for s in seqs]
#             seq_dict = {s.id: s for s in seqs}
#         elif os.path.exists(seqs):
#             seq_dict = None
#         ofile, cfile, = cdhit(seqs, out_file=out_file, temp_dir=temp_dir, threads=threads,
#                               threshold=threshold, make_db=False, quiet=quiet,
#                               max_memory=max_memory, debug=debug)
#         return parse_clusters(ofile, cfile, seq_dict=seq_dict, return_just_seq_ids=return_just_seq_ids)


# def cdhit(seqs, out_file=None, temp_dir=None, threshold=0.975, make_db=True,
#           quiet=False, threads=0, max_memory=800, retries=5, debug=False):
#     '''
#     Run CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that `abutils.utils.sequence.Sequence()`
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a file path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is `0.975`.

#         out_file (str): Path to the clustering output file. Default is to use
#             `tempfile.NamedTemporaryFile` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, the default
#             temporary directory is used (typically ``/tmp`` on MacOS and Linux).

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs).
#             Default is `True`.

#         quiet (bool): If `True`, surpresses printing of output/progress info. Default is `False`.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is `0`,
#             which results in all available cores being used.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the `-M` runtime option. Default is `800`.

#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty.
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are
#             not empty. Default is 5.

#         debug (bool): If `True`, print standard output and standard error from CD-HIT. Default is `False`.

#     Returns:

#         If `make_db` is `True`, returns the CD-HIT output file path, the CD-HIT cluster file path,
#             a `sqlite3` database connection object, and the database path. If `make_db` is `False`, only the
#             CD-HIT output file path and CD-HIT cluster file path are returned.
#     '''
#     start_time = time.time()
#     delete_input = False if debug else True
#     if not quiet:
#         print('CD-HIT: clustering {} seqeunces'.format(len(seqs)))
#     if out_file is None:
#         out_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#         out_file.close()
#         ofile = out_file.name
#     else:
#         ofile = os.path.expanduser(out_file)
#     cfile = ofile + '.clstr'
#     with open(ofile, 'w') as f: f.write('')
#     with open(cfile, 'w') as f: f.write('')
#     if isinstance(seqs, (list, tuple)):
#         seqs = [Sequence(s) for s in seqs]
#         ifile = _make_cdhit_input(seqs, temp_dir)
#     elif os.path.exists(seqs):
#         ifile = seqs
#         delete_input = False
#     word_size = _get_cdhit_wordsize(threshold)
#     cdhit_cmd = 'cdhit -i {} -o {} -c {} -n {} -d 0 -T {} -M {}'.format(ifile,
#                                                                         ofile,
#                                                                         threshold,
#                                                                         word_size,
#                                                                         threads,
#                                                                         max_memory)
#     while not all([os.path.getsize(cfile), os.path.getsize(cfile)]):
#         cluster = sp.Popen(cdhit_cmd,
#                         shell=True,
#                         stdout=sp.PIPE,
#                         stderr=sp.PIPE)
#         stdout, stderr = cluster.communicate()
#         if not retries:
#             break
#         retries -= 1
#     end_time = time.time()
#     if debug:
#         print(stdout)
#         print(stderr)
#     if delete_input:
#         os.unlink(ifile)
#     if not quiet:
#         print('CD-HIT: clustering took {:.2f} seconds'.format(end_time - start_time))
#     if make_db:
#         if not quiet:
#             print('CD-HIT: building a SQLite3 database')
#         seq_db, db_path = _build_seq_db(seqs, direc=temp_dir)
#         return ofile, cfile, seq_db, db_path
#     return ofile, cfile


# def parse_clusters(out_file, clust_file, seq_db=None, db_path=None, seq_dict=None, return_just_seq_ids=False):
#     '''
#     Parses CD-HIT output.

#     Args:

#         out_file (str): Path to the CD-HIT output file. Required.

#         clust_file (str): Path to the CD-HIT cluster file. Required.

#         seq_db (sqlite.Connection): SQLite3 `Connection` object. Default is ``None``. If not provided and
#             ``return_just_seq_ids`` is False, the returned ``Cluster`` objects will not contain any sequence
#             information beyond the sequence ID.

#         db_path (str): Path to a SQLite3 database file. Default is ``None``. Must be provided if
#             ``seq_db`` is also provided.

#         seq_dict (dict): A ``dict`` mapping sequence IDs to ``abutils.core.sequence.Sequence`` objects. Default
#             is ``None``. Typically used when a relatively small number of sequences are being clustered and
#             creating a ``sqlite3`` database would be overkill.

#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#     Returns:

#         A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     raw_clusters = [c.split('\n') for c in open(clust_file, 'r').read().split('\n>')]
#     if return_just_seq_ids:
#         ids = []
#         for rc in raw_clusters:
#             _ids = []
#             for c in rc[1:]:
#                 if c:
#                     _ids.append(c.split()[2][1:-3])
#             ids.append(_ids)
#         os.unlink(out_file)
#         os.unlink(clust_file)
#         return ids
#     os.unlink(out_file)
#     os.unlink(clust_file)
#     clusters = [Cluster(rc, seq_db, db_path, seq_dict) for rc in raw_clusters]
#     return CDHITResult(clusters, seq_db=seq_db, db_path=db_path, seq_dict=seq_dict)


# def _make_cdhit_input(seqs, temp_dir):
#     ifile = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#     ifile.close()
#     fastas = [s.fasta for s in seqs]
#     with open(ifile.name, 'w') as f:
#         f.write('\n'.join(fastas))
#     return ifile.name


# def _get_cdhit_wordsize(threshold):
#     if threshold > 0.7:
#         return 5
#     elif threshold > 0.6:
#         return 4
#     elif threshold > 0.5:
#         return 3
#     return 2


# def _build_seq_db(seqs, direc=None):
#     # '''
#     # Builds a SQLite3 database of sequences.

#     # Inputs are a list of Sequence objects and an optional directory to store the database.
#     # If ::direc:: is not provided, '/tmp' will be used.

#     # Returns a SQLite3 connection object and the database path.
#     # '''
#     direc = direc if direc is not None else '/tmp'
#     db_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
#     db_path = os.path.join(direc, db_name)
#     conn = sqlite3.connect(db_path)
#     c = conn.cursor()
#     create_cmd = '''CREATE TABLE seqs (id text, sequence text)'''
#     insert_cmd = 'INSERT INTO seqs VALUES (?,?)'
#     c.execute('DROP TABLE IF EXISTS seqs')
#     c.execute(create_cmd)
#     c.executemany(insert_cmd, [(str(s.id), str(s.sequence)) for s in seqs])
#     c.execute('CREATE INDEX seq_index ON seqs (id)')
#     return c, db_path
